@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -70,6 +70,12 @@ fz_trace_text_span(fz_context *ctx, fz_output *out, fz_text_span *span, int dept
 	int i;
 	fz_trace_indent(ctx, out, depth);
 	fz_write_printf(ctx, out, "<span font=\"%s\" wmode=\"%d\" bidi=\"%d\"", fz_font_name(ctx, span->font), span->wmode, span->bidi_level);
+	if (span->language != FZ_LANG_UNSET)
+	{
+		char text[8];
+		fz_string_from_text_language(text, span->language);
+		fz_write_printf(ctx, out, " lang=\"%s\"", text);
+	}
 	fz_write_printf(ctx, out, " trm=\"%g %g %g %g\">\n", span->trm.a, span->trm.b, span->trm.c, span->trm.d);
 	for (i = 0; i < span->len; i++)
 	{
@@ -546,7 +552,8 @@ fz_trace_begin_layer(fz_context *ctx, fz_device *dev_, const char *name)
 	fz_trace_device *dev = (fz_trace_device*)dev_;
 	fz_output *out = dev->out;
 	fz_trace_indent(ctx, out, dev->depth);
-	fz_write_printf(ctx, out, "<layer name=\"%s\"/>\n", name);
+	fz_write_printf(ctx, out, "<layer name=\"%s\">\n", name);
+	dev->depth++;
 }
 
 static void
@@ -554,8 +561,73 @@ fz_trace_end_layer(fz_context *ctx, fz_device *dev_)
 {
 	fz_trace_device *dev = (fz_trace_device*)dev_;
 	fz_output *out = dev->out;
+	dev->depth--;
 	fz_trace_indent(ctx, out, dev->depth);
-	fz_write_printf(ctx, out, "<end_layer/>\n");
+	fz_write_printf(ctx, out, "</layer>\n");
+}
+
+static void
+fz_trace_begin_structure(fz_context *ctx, fz_device *dev_, fz_structure standard, const char *raw, int uid)
+{
+	fz_trace_device *dev = (fz_trace_device*)dev_;
+	fz_output *out = dev->out;
+	const char *str = fz_structure_to_string(standard);
+	fz_trace_indent(ctx, out, dev->depth);
+	fz_write_printf(ctx, out, "<structure standard=\"%s\"", str);
+	if (raw && strcmp(str, raw))
+		fz_write_printf(ctx, out, " raw=\"%s\"", raw);
+	if (uid != 0)
+		fz_write_printf(ctx, out, " uid=\"%d\"", uid);
+	fz_write_printf(ctx, out, ">\n");
+	dev->depth++;
+}
+
+static void
+fz_trace_end_structure(fz_context *ctx, fz_device *dev_)
+{
+	fz_trace_device *dev = (fz_trace_device*)dev_;
+	fz_output *out = dev->out;
+	dev->depth--;
+	fz_trace_indent(ctx, out, dev->depth);
+	fz_write_printf(ctx, out, "</structure>\n");
+}
+
+static const char *
+metatext_type(fz_metatext meta)
+{
+	switch (meta)
+	{
+	case FZ_METATEXT_ABBREVIATION:
+		return "abbreviation";
+	case FZ_METATEXT_ACTUALTEXT:
+		return "actualtext";
+	case FZ_METATEXT_ALT:
+		return "alt";
+	case FZ_METATEXT_TITLE:
+		return "title";
+	}
+	return "????";
+}
+
+static void
+fz_trace_begin_metatext(fz_context *ctx, fz_device *dev_, fz_metatext meta, const char *txt)
+{
+	fz_trace_device *dev = (fz_trace_device*)dev_;
+	fz_output *out = dev->out;
+	const char *type = metatext_type(meta);
+	fz_trace_indent(ctx, out, dev->depth);
+	fz_write_printf(ctx, out, "<metatext type=\"%s\" txt=\"%s\">\n", type, txt ? txt : "");
+	dev->depth++;
+}
+
+static void
+fz_trace_end_metatext(fz_context *ctx, fz_device *dev_)
+{
+	fz_trace_device *dev = (fz_trace_device*)dev_;
+	fz_output *out = dev->out;
+	dev->depth--;
+	fz_trace_indent(ctx, out, dev->depth);
+	fz_write_printf(ctx, out, "</metatext>\n");
 }
 
 static void
@@ -613,6 +685,12 @@ fz_device *fz_new_trace_device(fz_context *ctx, fz_output *out)
 	dev->super.begin_layer = fz_trace_begin_layer;
 	dev->super.end_layer = fz_trace_end_layer;
 
+	dev->super.begin_structure = fz_trace_begin_structure;
+	dev->super.end_structure = fz_trace_end_structure;
+
+	dev->super.begin_metatext = fz_trace_begin_metatext;
+	dev->super.end_metatext = fz_trace_end_metatext;
+
 	dev->super.render_flags = fz_trace_render_flags;
 	dev->super.set_default_colorspaces = fz_trace_set_default_colorspaces;
 
