@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -262,16 +262,6 @@ pdf_annot_request_resynthesis(fz_context *ctx, pdf_annot *annot)
 	if (annot == NULL)
 		return;
 
-	/* Some appearances can NEVER be resynthesised. Spot those here. */
-	if (pdf_name_eq(ctx, pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)), PDF_NAME(Stamp)))
-	{
-		/* We can't resynthesise a stamp if we don't have the name! */
-		/* In particular, this is the case we hit when we are asked to
-		 * resynthesise an e-sig. Just exit and do nothing. */
-		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Name)) == NULL)
-			return;
-	}
-
 	/* If there are no changes, there is no need to request a resynthesis
 	 * (and indeed, we must not, because any changes caused by the resynth
 	 * will go in as implicit changes into a potentially-non-existent
@@ -565,7 +555,7 @@ pdf_create_link(fz_context *ctx, pdf_page *page, fz_rect bbox, const char *uri)
 		ind_obj = pdf_new_indirect(ctx, doc, ind_obj_num, 0);
 		pdf_array_push(ctx, annot_arr, ind_obj);
 
-		link = fz_new_link(ctx, bbox, uri);
+		link = (fz_link *) pdf_new_link(ctx, page, bbox, uri, annot_obj);
 
 		linkp = &page->links;
 
@@ -589,6 +579,47 @@ pdf_create_link(fz_context *ctx, pdf_page *page, fz_rect bbox, const char *uri)
 	return fz_keep_link(ctx, link);
 }
 
+void pdf_delete_link(fz_context *ctx, pdf_page *page, fz_link *link)
+{
+	fz_link **linkptr;
+	pdf_obj *annots;
+	int i;
+
+	if (link == NULL || page == NULL || page != ((pdf_link *) link)->page)
+		return;
+
+	for (linkptr = &page->links; *linkptr; linkptr = &((*linkptr)->next))
+	{
+		if (*linkptr == link)
+			break;
+	}
+
+	if (*linkptr == NULL)
+		return;
+
+	pdf_begin_operation(ctx, page->doc, "Delete Link");
+
+	fz_try(ctx)
+	{
+
+		annots = pdf_dict_get(ctx, page->obj, PDF_NAME(Annots));
+		i = pdf_array_find(ctx, annots, ((pdf_link *) link)->obj);
+		if (i >= 0)
+			pdf_array_delete(ctx, annots, i);
+		*linkptr = link->next;
+		link->next = NULL;
+		fz_drop_link(ctx, link);
+	}
+	fz_always(ctx)
+	{
+		pdf_end_operation(ctx, page->doc);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+}
+
 static pdf_obj *
 pdf_add_popup_annot(fz_context *ctx, pdf_annot *annot)
 {
@@ -694,6 +725,7 @@ pdf_create_annot(fz_context *ctx, pdf_page *page, enum pdf_annot_type type)
 				fz_rect stamp_rect = { 12, 12, 12+190, 12+50 };
 				pdf_set_annot_rect(ctx, annot, stamp_rect);
 				pdf_set_annot_color(ctx, annot, 3, red);
+				pdf_set_annot_icon_name(ctx, annot, "Draft");
 			}
 			break;
 
@@ -868,12 +900,13 @@ pdf_delete_annot(fz_context *ctx, pdf_page *page, pdf_annot *annot)
 
 		/* The garbage collection pass when saving will remove the annot object,
 		 * removing it here may break files if multiple pages use the same annot. */
-
-		/* And free it. */
-		pdf_drop_annot(ctx, annot);
 	}
 	fz_always(ctx)
+	{
+		/* Drop the reference to annot previously held by the page list. */
+		pdf_drop_annot(ctx, annot);
 		pdf_end_operation(ctx, page->doc);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
@@ -929,13 +962,50 @@ pdf_set_annot_flags(fz_context *ctx, pdf_annot *annot, int flags)
 	pdf_dirty_annot(ctx, annot);
 }
 
+static pdf_obj *rect_subtypes[] = {
+	PDF_NAME(Text),
+	PDF_NAME(FreeText),
+	PDF_NAME(Square),
+	PDF_NAME(Circle),
+	PDF_NAME(Redact),
+	PDF_NAME(Stamp),
+	PDF_NAME(Caret),
+	PDF_NAME(Popup),
+	PDF_NAME(FileAttachment),
+	PDF_NAME(Sound),
+	PDF_NAME(Movie),
+	PDF_NAME(Widget),
+	NULL,
+};
+
+int
+pdf_annot_has_rect(fz_context *ctx, pdf_annot *annot)
+{
+	/* True for annotations where the user can manipulate the size or location
+	 * of the annotation through the Rect.
+	 * False for annotations where the Rect is computed from other
+	 * annotation data such as InkList, QuadPoints, and Vertices.
+	 */
+	return is_allowed_subtype_wrap(ctx, annot, PDF_NAME(Rect), rect_subtypes);
+}
+
 fz_rect
 pdf_annot_rect(fz_context *ctx, pdf_annot *annot)
 {
 	fz_matrix page_ctm;
 	fz_rect annot_rect;
-	pdf_page_transform(ctx, annot->page, NULL, &page_ctm);
-	annot_rect = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
+
+	pdf_annot_push_local_xref(ctx, annot);
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(Rect), rect_subtypes);
+		pdf_page_transform(ctx, annot->page, NULL, &page_ctm);
+		annot_rect = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 	return fz_transform_rect(annot_rect, page_ctm);
 }
 
@@ -948,6 +1018,8 @@ pdf_set_annot_rect(fz_context *ctx, pdf_annot *annot, fz_rect rect)
 
 	fz_try(ctx)
 	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(Rect), rect_subtypes);
+
 		pdf_page_transform(ctx, annot->page, NULL, &page_ctm);
 		inv_page_ctm = fz_invert_matrix(page_ctm);
 		rect = fz_transform_rect(rect, inv_page_ctm);
@@ -1091,7 +1163,7 @@ pdf_annot_icon_name(fz_context *ctx, pdf_annot *annot)
 			}
 			if (pdf_name_eq(ctx, subtype, PDF_NAME(Stamp)))
 			{
-				ret = "Draft";
+				ret = ""; // Should be "Draft" according to spec
 				break;
 			}
 			if (pdf_name_eq(ctx, subtype, PDF_NAME(FileAttachment)))
@@ -1123,7 +1195,10 @@ pdf_set_annot_icon_name(fz_context *ctx, pdf_annot *annot, const char *name)
 	fz_try(ctx)
 	{
 		check_allowed_subtypes(ctx, annot, PDF_NAME(Name), icon_name_subtypes);
-		pdf_dict_put_name(ctx, annot->obj, PDF_NAME(Name), name);
+		if (name)
+			pdf_dict_put_name(ctx, annot->obj, PDF_NAME(Name), name);
+		else
+			pdf_dict_del(ctx, annot->obj, PDF_NAME(Name));
 	}
 	fz_always(ctx)
 		end_annot_op(ctx, annot);
@@ -1133,6 +1208,27 @@ pdf_set_annot_icon_name(fz_context *ctx, pdf_annot *annot, const char *name)
 	pdf_dirty_annot(ctx, annot);
 }
 
+int
+pdf_annot_is_standard_stamp(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *name = pdf_dict_get(ctx, annot->obj, PDF_NAME(Name));
+	if (pdf_name_eq(ctx, name, PDF_NAME(Approved))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(AsIs))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Confidential))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Departmental))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Draft))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Experimental))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Expired))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Final))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(ForComment))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(ForPublicRelease))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(NotApproved))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(NotForPublicRelease))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(Sold))) return 1;
+	else if (pdf_name_eq(ctx, name, PDF_NAME(TopSecret))) return 1;
+	else return 0;
+}
+
 enum pdf_line_ending pdf_line_ending_from_name(fz_context *ctx, pdf_obj *end)
 {
 	if (pdf_name_eq(ctx, end, PDF_NAME(None))) return PDF_ANNOT_LE_NONE;
@@ -1330,10 +1426,355 @@ pdf_set_annot_border(fz_context *ctx, pdf_annot *annot, float w)
 		pdf_obj *bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
 		if (!pdf_is_dict(ctx, bs))
 			bs = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BS), 1);
+		pdf_dict_put(ctx, bs, PDF_NAME(Type), PDF_NAME(Border));
 		pdf_dict_put_real(ctx, bs, PDF_NAME(W), w);
 
 		pdf_dict_del(ctx, annot->obj, PDF_NAME(Border)); /* deprecated */
-		pdf_dict_del(ctx, annot->obj, PDF_NAME(BE)); /* not supported */
+		pdf_dict_del(ctx, annot->obj, PDF_NAME(BE)); /* no effect */
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+static pdf_obj *border_style_subtypes[] = {
+	PDF_NAME(Circle),
+	PDF_NAME(FreeText),
+	PDF_NAME(Ink),
+	PDF_NAME(Line),
+	PDF_NAME(Polygon),
+	PDF_NAME(PolyLine),
+	PDF_NAME(Square),
+	NULL,
+};
+
+static pdf_obj *border_effect_subtypes[] = {
+	PDF_NAME(Circle),
+	PDF_NAME(FreeText),
+	PDF_NAME(Polygon),
+	PDF_NAME(Square),
+	NULL,
+};
+
+int
+pdf_annot_has_border(fz_context *ctx, pdf_annot *annot)
+{
+	return is_allowed_subtype_wrap(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+}
+
+int
+pdf_annot_has_border_effect(fz_context *ctx, pdf_annot *annot)
+{
+	return is_allowed_subtype_wrap(ctx, annot, PDF_NAME(BE), border_effect_subtypes);
+}
+
+enum pdf_border_style
+pdf_annot_border_style(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *bs, *s;
+	enum pdf_border_style style;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		s = pdf_dict_get(ctx, bs, PDF_NAME(S));
+
+		if (s == PDF_NAME(D))
+			style = PDF_BORDER_STYLE_DASHED;
+		else if (s == PDF_NAME(B))
+			style = PDF_BORDER_STYLE_BEVELED;
+		else if (s == PDF_NAME(I))
+			style = PDF_BORDER_STYLE_INSET;
+		else if (s == PDF_NAME(U))
+			style = PDF_BORDER_STYLE_UNDERLINE;
+		else
+			style = PDF_BORDER_STYLE_SOLID;
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return style;
+}
+
+float
+pdf_annot_border_width(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *border, *bs, *w;
+	float width;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+
+		/* if values missing, fall back to deprecated /Border array */
+		border = pdf_dict_get(ctx, annot->obj, PDF_NAME(Border));
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		w = pdf_dict_get(ctx, bs, PDF_NAME(W));
+		if (!pdf_is_number(ctx, w) && pdf_is_dict(ctx, border))
+			w = pdf_array_get(ctx, border, 2);
+		width = pdf_to_real(ctx, w);
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return width;
+}
+
+int
+pdf_annot_border_dash_count(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *bs, *d;
+	int count;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		d = pdf_dict_get(ctx, bs, PDF_NAME(D));
+		count = pdf_array_len(ctx, d);
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return count;
+}
+
+float
+pdf_annot_border_dash_item(fz_context *ctx, pdf_annot *annot, int i)
+{
+	pdf_obj *bs, *d;
+	float length;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		d = pdf_dict_get(ctx, bs, PDF_NAME(D));
+		length = pdf_array_get_real(ctx, d, i);
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return length;
+}
+
+enum pdf_border_effect
+pdf_annot_border_effect(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *be;
+	enum pdf_border_effect effect;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BE), border_effect_subtypes);
+		be = pdf_dict_get(ctx, annot->obj, PDF_NAME(BE));
+		if (pdf_dict_get(ctx, be, PDF_NAME(S)) == PDF_NAME(C))
+			effect = PDF_BORDER_EFFECT_CLOUDY;
+		else
+			effect = PDF_BORDER_EFFECT_NONE;
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return effect;
+}
+
+float
+pdf_annot_border_effect_intensity(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *be;
+	float intensity;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BE), border_effect_subtypes);
+		be = pdf_dict_get(ctx, annot->obj, PDF_NAME(BE));
+		intensity = pdf_dict_get_real(ctx, be, PDF_NAME(I));
+	}
+	fz_always(ctx)
+		pdf_annot_pop_local_xref(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return intensity;
+}
+
+void
+pdf_set_annot_border_width(fz_context *ctx, pdf_annot *annot, float width)
+{
+	pdf_obj *bs;
+
+	begin_annot_op(ctx, annot, "Set border width");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		if (!pdf_is_dict(ctx, bs))
+			bs = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BS), 1);
+		pdf_dict_put(ctx, bs, PDF_NAME(Type), PDF_NAME(Border));
+		pdf_dict_put_real(ctx, bs, PDF_NAME(W), width);
+		pdf_dict_del(ctx, annot->obj, PDF_NAME(Border)); /* deprecated */
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+void
+pdf_set_annot_border_style(fz_context *ctx, pdf_annot *annot, enum pdf_border_style style)
+{
+	pdf_obj *bs, *s;
+
+	begin_annot_op(ctx, annot, "Set border style");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		if (!pdf_is_dict(ctx, bs))
+			bs = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BS), 1);
+		pdf_dict_put(ctx, bs, PDF_NAME(Type), PDF_NAME(Border));
+		switch (style)
+		{
+		default:
+		case PDF_BORDER_STYLE_SOLID: s = PDF_NAME(S); break;
+		case PDF_BORDER_STYLE_DASHED: s = PDF_NAME(D); break;
+		case PDF_BORDER_STYLE_BEVELED: s = PDF_NAME(B); break;
+		case PDF_BORDER_STYLE_INSET: s = PDF_NAME(I); break;
+		case PDF_BORDER_STYLE_UNDERLINE: s = PDF_NAME(U); break;
+		}
+		pdf_dict_put(ctx, bs, PDF_NAME(S), s);
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+void
+pdf_clear_annot_border_dash(fz_context *ctx, pdf_annot *annot)
+{
+	pdf_obj *bs;
+
+	begin_annot_op(ctx, annot, "Clear border dash pattern");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		if (!pdf_is_dict(ctx, bs))
+			bs = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BS), 1);
+		pdf_dict_del(ctx, bs, PDF_NAME(D));
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+void
+pdf_add_annot_border_dash_item(fz_context *ctx, pdf_annot *annot, float length)
+{
+	pdf_obj *bs, *d;
+
+	begin_annot_op(ctx, annot, "Add border dash pattern item");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BS), border_style_subtypes);
+		bs = pdf_dict_get(ctx, annot->obj, PDF_NAME(BS));
+		if (!pdf_is_dict(ctx, bs))
+			bs = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BS), 1);
+		d = pdf_dict_get(ctx, bs, PDF_NAME(D));
+		if (!pdf_is_array(ctx, d))
+			d = pdf_dict_put_array(ctx, bs, PDF_NAME(D), 1);
+		pdf_array_push_real(ctx, d, length);
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+void
+pdf_set_annot_border_effect(fz_context *ctx, pdf_annot *annot, enum pdf_border_effect effect)
+{
+	pdf_obj *be, *s;
+
+	begin_annot_op(ctx, annot, "Set border effect");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BE), border_effect_subtypes);
+		be = pdf_dict_get(ctx, annot->obj, PDF_NAME(BE));
+		if (!pdf_is_dict(ctx, be))
+			be = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BE), 1);
+		switch (effect)
+		{
+		default:
+		case PDF_BORDER_EFFECT_NONE: s = PDF_NAME(S); break;
+		case PDF_BORDER_EFFECT_CLOUDY: s = PDF_NAME(C); break;
+		}
+		pdf_dict_put(ctx, be, PDF_NAME(S), s);
+	}
+	fz_always(ctx)
+		end_annot_op(ctx, annot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	pdf_dirty_annot(ctx, annot);
+}
+
+void
+pdf_set_annot_border_effect_intensity(fz_context *ctx, pdf_annot *annot, float intensity)
+{
+	pdf_obj *be;
+
+	begin_annot_op(ctx, annot, "Set border effect intensity");
+
+	fz_try(ctx)
+	{
+		check_allowed_subtypes(ctx, annot, PDF_NAME(BE), border_effect_subtypes);
+		be = pdf_dict_get(ctx, annot->obj, PDF_NAME(BE));
+		if (!pdf_is_dict(ctx, be))
+			be = pdf_dict_put_dict(ctx, annot->obj, PDF_NAME(BE), 1);
+		pdf_dict_put_real(ctx, be, PDF_NAME(I), intensity);
 	}
 	fz_always(ctx)
 		end_annot_op(ctx, annot);
@@ -2261,8 +2702,6 @@ pdf_set_annot_ink_list(fz_context *ctx, pdf_annot *annot, int n, const int *coun
 		pdf_page_transform(ctx, annot->page, NULL, &page_ctm);
 		inv_page_ctm = fz_invert_matrix(page_ctm);
 
-		// TODO: update Rect (in update appearance perhaps?)
-
 		ink_list = pdf_new_array(ctx, doc, n);
 		for (i = 0; i < n; ++i)
 		{
@@ -2831,6 +3270,48 @@ pdf_set_annot_appearance_from_display_list(fz_context *ctx, pdf_annot *annot, co
 		fz_rethrow(ctx);
 }
 
+void pdf_set_annot_stamp_image(fz_context *ctx, pdf_annot *annot, fz_image *img)
+{
+	pdf_document *doc = annot->page->doc;
+	fz_buffer *buf = NULL;
+	pdf_obj *res = NULL;
+	pdf_obj *res_xobj;
+
+	begin_annot_op(ctx, annot, "Set stamp image");
+
+	fz_var(res);
+	fz_var(buf);
+
+	fz_try(ctx)
+	{
+		// Shrink Rect to fit image, maintaining aspect ratio.
+		fz_rect rect = pdf_bound_annot(ctx, annot);
+		float s = fz_min((rect.x1 - rect.x0) / img->w, (rect.y1 - rect.y0) / img->h);
+		rect.x1 = rect.x0 + img->w * s;
+		rect.y1 = rect.y0 + img->h * s;
+
+		// Add image resource
+		res = pdf_add_new_dict(ctx, doc, 1);
+		res_xobj = pdf_dict_put_dict(ctx, res, PDF_NAME(XObject), 1);
+		pdf_dict_put_drop(ctx, res_xobj, PDF_NAME(I), pdf_add_image(ctx, doc, img));
+
+		buf = fz_new_buffer_from_shared_data(ctx, (const unsigned char*)"/I Do\n", 6);
+
+		pdf_set_annot_appearance(ctx, annot, "N", NULL, fz_identity, fz_unit_rect, res, buf);
+		pdf_set_annot_rect(ctx, annot, rect);
+	}
+	fz_always(ctx)
+	{
+		fz_drop_buffer(ctx, buf);
+		pdf_drop_obj(ctx, res);
+		end_annot_op(ctx, annot);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+}
+
 static pdf_obj *filespec_subtypes[] = {
 	PDF_NAME(FileAttachment),
 	NULL,
@@ -2872,8 +3353,8 @@ pdf_set_annot_filespec(fz_context *ctx, pdf_annot *annot, pdf_obj *fs)
 
 	fz_try(ctx)
 	{
-		check_allowed_subtypes(ctx, annot, PDF_NAME(M), markup_subtypes);
-		pdf_dict_put_drop(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(FS), fs);
+		check_allowed_subtypes(ctx, annot, PDF_NAME(M), filespec_subtypes);
+		pdf_dict_put(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(FS), fs);
 	}
 	fz_always(ctx)
 		end_annot_op(ctx, annot);
@@ -2882,3 +3363,15 @@ pdf_set_annot_filespec(fz_context *ctx, pdf_annot *annot, pdf_obj *fs)
 
 	pdf_dirty_annot(ctx, annot);
 }
+
+int
+pdf_annot_hidden_for_editing(fz_context *ctx, pdf_annot *annot)
+{
+	return annot->hidden_editing;
+}
+
+void
+pdf_set_annot_hidden_for_editing(fz_context *ctx, pdf_annot *annot, int hidden)
+{
+	annot->hidden_editing = hidden;
+}
