@@ -27,18 +27,18 @@
 #include <assert.h>
 
 static void
-pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *xobj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up);
+pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *xobj, pdf_obj *page_res, pdf_filter_options *options, pdf_cycle_list *cycle_up);
 
 static void
-pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up);
+pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *options, pdf_cycle_list *cycle_up);
 
 static void
-pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_obj *res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
+pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_obj *res, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
 	pdf_obj *obj;
 	int i, n;
 
-	if (!filter->recurse)
+	if (!options->recurse)
 		return;
 
 	/* ExtGState */
@@ -55,7 +55,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				if (g)
 				{
 					/* Transparency group XObject */
-					pdf_filter_xobject(ctx, doc, g, in_res, filter, cycle_up);
+					pdf_filter_xobject(ctx, doc, g, in_res, options, cycle_up);
 				}
 			}
 		}
@@ -71,13 +71,13 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *pat = pdf_dict_get_val(ctx, obj, i);
 			if (pat && pdf_dict_get_int(ctx, pat, PDF_NAME(PatternType)) == 1)
 			{
-				pdf_filter_xobject(ctx, doc, pat, in_res, filter, cycle_up);
+				pdf_filter_xobject(ctx, doc, pat, in_res, options, cycle_up);
 			}
 		}
 	}
 
 	/* XObject */
-	if (!filter->instance_forms)
+	if (!options->instance_forms)
 	{
 		obj = pdf_dict_get(ctx, res, PDF_NAME(XObject));
 		if (obj)
@@ -88,7 +88,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				pdf_obj *xobj = pdf_dict_get_val(ctx, obj, i);
 				if (xobj && pdf_dict_get(ctx, xobj, PDF_NAME(Subtype)) == PDF_NAME(Form))
 				{
-					pdf_filter_xobject(ctx, doc, xobj, in_res, filter, cycle_up);
+					pdf_filter_xobject(ctx, doc, xobj, in_res, options, cycle_up);
 				}
 			}
 		}
@@ -104,7 +104,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *font = pdf_dict_get_val(ctx, obj, i);
 			if (font && pdf_dict_get(ctx, font, PDF_NAME(Subtype)) == PDF_NAME(Type3))
 			{
-				pdf_filter_type3(ctx, doc, font, in_res, filter, cycle_up);
+				pdf_filter_type3(ctx, doc, font, in_res, options, cycle_up);
 			}
 		}
 	}
@@ -130,45 +130,50 @@ pdf_filter_content_stream(
 	pdf_obj *in_stm,
 	pdf_obj *in_res,
 	fz_matrix transform,
-	pdf_filter_options *filter,
+	pdf_filter_options *options,
 	int struct_parents,
 	fz_buffer **out_buf,
 	pdf_obj **out_res,
 	pdf_cycle_list *cycle_up)
 {
 	pdf_processor *proc_buffer = NULL;
-	pdf_processor *proc_filter = NULL;
+	pdf_processor *top = NULL;
+	pdf_processor **list = NULL;
+	int num_filters = 0;
+	int i;
 
 	fz_var(proc_buffer);
-	fz_var(proc_filter);
 
 	*out_buf = NULL;
 	*out_res = NULL;
 
+	if (options->filters)
+		for (; options->filters[num_filters].filter != NULL; num_filters++);
+
+	if (num_filters > 0)
+		list = fz_calloc(ctx, num_filters, sizeof(pdf_processor *));
+
 	fz_try(ctx)
 	{
 		*out_buf = fz_new_buffer(ctx, 1024);
-		proc_buffer = pdf_new_buffer_processor(ctx, *out_buf, filter->ascii);
-		if (filter->sanitize)
+		top = proc_buffer = pdf_new_buffer_processor(ctx, *out_buf, options->ascii);
+		if (num_filters > 0)
 		{
-			*out_res = pdf_new_dict(ctx, doc, 1);
-			proc_filter = pdf_new_filter_processor(ctx, doc, proc_buffer, in_res, *out_res, struct_parents, transform, filter);
-			pdf_process_contents(ctx, proc_filter, doc, in_res, in_stm, NULL);
-			pdf_close_processor(ctx, proc_filter);
+			for (i = num_filters - 1; i >= 0; i--)
+				top = list[i] = options->filters[i].filter(ctx, doc, top, struct_parents, transform, options, options->filters[i].options);
 		}
-		else
-		{
-			*out_res = pdf_keep_obj(ctx, in_res);
-			pdf_process_contents(ctx, proc_buffer, doc, in_res, in_stm, NULL);
-		}
-		pdf_close_processor(ctx, proc_buffer);
 
-		pdf_filter_resources(ctx, doc, in_res, *out_res, filter, cycle_up);
+		pdf_process_contents(ctx, top, doc, in_res, in_stm, NULL, out_res);
+		pdf_close_processor(ctx, top);
+
+		pdf_filter_resources(ctx, doc, in_res, *out_res, options, cycle_up);
 	}
 	fz_always(ctx)
 	{
-		pdf_drop_processor(ctx, proc_filter);
+		for (i = 0; i < num_filters; i++)
+			pdf_drop_processor(ctx, list[i]);
 		pdf_drop_processor(ctx, proc_buffer);
+		fz_free(ctx, list);
 	}
 	fz_catch(ctx)
 	{
@@ -186,7 +191,7 @@ pdf_filter_content_stream(
 	shared between all off the CharProcs.
 */
 static void
-pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
+pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page_res, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
 	pdf_cycle_list cycle;
 	pdf_processor *proc_buffer = NULL;
@@ -195,10 +200,19 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 	pdf_obj *out_res = NULL;
 	pdf_obj *charprocs;
 	int i, n;
+	int num_filters = 0;
+	pdf_processor **list = NULL;
+	fz_buffer *buffer = NULL;
+	pdf_processor *top = NULL;
+	pdf_obj *res = NULL;
+	fz_buffer *new_buf = NULL;
 
 	fz_var(out_res);
 	fz_var(proc_buffer);
 	fz_var(proc_filter);
+	fz_var(buffer);
+	fz_var(res);
+	fz_var(new_buf);
 
 	/* We cannot combine instancing with type3 fonts. The new names for
 	 * instanced form/image resources would clash, since they start over for
@@ -206,75 +220,74 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 	 * use instancing with redaction, and redaction doesn't clean type3
 	 * fonts.
 	 */
-	assert(!filter->instance_forms);
+	assert(!options->instance_forms);
 
 	/* Avoid recursive cycles! */
 	if (pdf_cycle(ctx, &cycle, cycle_up, obj))
 		return;
 
+	if (options->filters)
+		for (; options->filters[num_filters].filter != NULL; num_filters++);
+
+	if (num_filters > 0)
+		list = fz_calloc(ctx, num_filters, sizeof(pdf_processor *));
+
 	fz_try(ctx)
 	{
 		in_res = pdf_dict_get(ctx, obj, PDF_NAME(Resources));
 		if (!in_res)
 			in_res = page_res;
 
-		if (filter->sanitize)
-			out_res = pdf_new_dict(ctx, doc, 1);
-		else
-			out_res = pdf_keep_obj(ctx, in_res);
+		buffer = fz_new_buffer(ctx, 1024);
+		top = proc_buffer = pdf_new_buffer_processor(ctx, buffer, options->ascii);
+		if (num_filters > 0)
+		{
+			for (i = num_filters - 1; i >= 0; i--)
+				top = list[i] = options->filters[i].filter(ctx, doc, top, -1, fz_identity, options, options->filters[i].options);
+		}
 
+		pdf_processor_push_resources(ctx, top, in_res);
 		charprocs = pdf_dict_get(ctx, obj, PDF_NAME(CharProcs));
 		n = pdf_dict_len(ctx, charprocs);
 		for (i = 0; i < n; i++)
 		{
 			pdf_obj *val = pdf_dict_get_val(ctx, charprocs, i);
-			fz_buffer *buffer = fz_new_buffer(ctx, 1024);
-			fz_try(ctx)
-			{
-				proc_buffer = pdf_new_buffer_processor(ctx, buffer, filter->ascii);
-				if (filter->sanitize)
-				{
-					proc_filter = pdf_new_filter_processor(ctx, doc, proc_buffer, in_res, out_res, -1, fz_identity, filter);
-					pdf_process_contents(ctx, proc_filter, doc, in_res, val, NULL);
-					pdf_close_processor(ctx, proc_filter);
-				}
-				else
-				{
-					pdf_process_contents(ctx, proc_buffer, doc, in_res, val, NULL);
-				}
-				pdf_close_processor(ctx, proc_buffer);
 
-				pdf_update_stream(ctx, doc, val, buffer, 0);
-			}
-			fz_always(ctx)
+			fz_clear_buffer(ctx, buffer);
+			pdf_process_raw_contents(ctx, top, doc, in_res, val, NULL);
+
+			pdf_close_processor(ctx, proc_buffer);
+
+			if (!options->no_update)
 			{
-				pdf_drop_processor(ctx, proc_filter);
-				pdf_drop_processor(ctx, proc_buffer);
-				fz_drop_buffer(ctx, buffer);
-			}
-			fz_catch(ctx)
-			{
-				fz_rethrow(ctx);
+				new_buf = fz_clone_buffer(ctx, buffer);
+				pdf_update_stream(ctx, doc, val, new_buf, 0);
+				fz_drop_buffer(ctx, new_buf);
+				new_buf = NULL;
 			}
 		}
 
-		pdf_filter_resources(ctx, doc, in_res, out_res, filter, &cycle);
-
-		if (filter->sanitize)
-			pdf_dict_put(ctx, obj, PDF_NAME(Resources), out_res);
 	}
 	fz_always(ctx)
 	{
-		pdf_drop_obj(ctx, out_res);
+		res = pdf_processor_pop_resources(ctx, top);
+		for (i = 0; i < num_filters; i++)
+			pdf_drop_processor(ctx, list[i]);
+		pdf_drop_processor(ctx, proc_buffer);
+		fz_free(ctx, list);
+		fz_drop_buffer(ctx, new_buf);
+		fz_drop_buffer(ctx, buffer);
 	}
 	fz_catch(ctx)
 	{
+		pdf_drop_obj(ctx, res);
 		fz_rethrow(ctx);
 	}
+	pdf_dict_put_drop(ctx, obj, PDF_NAME(Resources), res);
 }
 
 static void
-pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *page_res, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
+pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *page_res, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
 	pdf_cycle_list cycle;
 	pdf_obj *struct_parents_obj;
@@ -304,9 +317,12 @@ pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *pa
 		return;
 	fz_try(ctx)
 	{
-		pdf_filter_content_stream(ctx, doc, stm, old_res, fz_identity, filter, struct_parents, &new_buf, &new_res, &cycle);
-		pdf_update_stream(ctx, doc, stm, new_buf, 0);
-		pdf_dict_put(ctx, stm, PDF_NAME(Resources), new_res);
+		pdf_filter_content_stream(ctx, doc, stm, old_res, fz_identity, options, struct_parents, &new_buf, &new_res, &cycle);
+		if (!options->no_update)
+		{
+			pdf_update_stream(ctx, doc, stm, new_buf, 0);
+			pdf_dict_put(ctx, stm, PDF_NAME(Resources), new_res);
+		}
 	}
 	fz_always(ctx)
 	{
@@ -318,7 +334,7 @@ pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *pa
 }
 
 pdf_obj *
-pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_res, fz_matrix transform, pdf_filter_options *filter, pdf_cycle_list *cycle_up)
+pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_res, fz_matrix transform, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
 	pdf_cycle_list cycle;
 	pdf_document *doc = pdf_get_bound_document(ctx, old_xobj);
@@ -354,9 +370,12 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 	fz_try(ctx)
 	{
 		new_xobj = pdf_add_object_drop(ctx, doc, pdf_copy_dict(ctx, old_xobj));
-		pdf_filter_content_stream(ctx, doc, old_xobj, old_res, transform, filter, struct_parents, &new_buf, &new_res, &cycle);
-		pdf_update_stream(ctx, doc, new_xobj, new_buf, 0);
-		pdf_dict_put(ctx, new_xobj, PDF_NAME(Resources), new_res);
+		pdf_filter_content_stream(ctx, doc, old_xobj, old_res, transform, options, struct_parents, &new_buf, &new_res, &cycle);
+		if (!options->no_update)
+		{
+			pdf_update_stream(ctx, doc, new_xobj, new_buf, 0);
+			pdf_dict_put(ctx, new_xobj, PDF_NAME(Resources), new_res);
+		}
 	}
 	fz_always(ctx)
 	{
@@ -372,7 +391,7 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 	return new_xobj;
 }
 
-void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_filter_options *filter)
+void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_filter_options *options)
 {
 	pdf_obj *contents, *old_res;
 	pdf_obj *struct_parents_obj;
@@ -388,21 +407,24 @@ void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page
 	contents = pdf_page_contents(ctx, page);
 	old_res = pdf_page_resources(ctx, page);
 
-	pdf_filter_content_stream(ctx, doc, contents, old_res, fz_identity, filter, struct_parents, &buffer, &new_res, NULL);
+	pdf_filter_content_stream(ctx, doc, contents, old_res, fz_identity, options, struct_parents, &buffer, &new_res, NULL);
 
 	fz_try(ctx)
 	{
-		if (filter->end_page)
-			filter->end_page(ctx, buffer, filter->opaque);
-		/* If contents is not a stream it's an array of streams or missing. */
-		if (!pdf_is_stream(ctx, contents))
+		if (options->complete)
+			options->complete(ctx, buffer, options->opaque);
+		if (!options->no_update)
 		{
-			/* Create a new stream object to replace the array of streams or missing object. */
-			contents = pdf_add_object_drop(ctx, doc, pdf_new_dict(ctx, doc, 1));
-			pdf_dict_put_drop(ctx, page->obj, PDF_NAME(Contents), contents);
+			/* If contents is not a stream it's an array of streams or missing. */
+			if (!pdf_is_stream(ctx, contents))
+			{
+				/* Create a new stream object to replace the array of streams or missing object. */
+				contents = pdf_add_object_drop(ctx, doc, pdf_new_dict(ctx, doc, 1));
+				pdf_dict_put_drop(ctx, page->obj, PDF_NAME(Contents), contents);
+			}
+			pdf_update_stream(ctx, doc, contents, buffer, 0);
+			pdf_dict_put(ctx, page->obj, PDF_NAME(Resources), new_res);
 		}
-		pdf_update_stream(ctx, doc, contents, buffer, 0);
-		pdf_dict_put(ctx, page->obj, PDF_NAME(Resources), new_res);
 	}
 	fz_always(ctx)
 	{
@@ -413,7 +435,7 @@ void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page
 		fz_rethrow(ctx);
 }
 
-void pdf_filter_annot_contents(fz_context *ctx, pdf_document *doc, pdf_annot *annot, pdf_filter_options *filter)
+void pdf_filter_annot_contents(fz_context *ctx, pdf_document *doc, pdf_annot *annot, pdf_filter_options *options)
 {
 	pdf_obj *ap = pdf_dict_get(ctx, annot->obj, PDF_NAME(AP));
 	if (pdf_is_dict(ctx, ap))
@@ -424,7 +446,7 @@ void pdf_filter_annot_contents(fz_context *ctx, pdf_document *doc, pdf_annot *an
 			pdf_obj *stm = pdf_dict_get_val(ctx, ap, i);
 			if (pdf_is_stream(ctx, stm))
 			{
-				pdf_filter_xobject(ctx, doc, stm, NULL, filter, NULL);
+				pdf_filter_xobject(ctx, doc, stm, NULL, options, NULL);
 			}
 		}
 	}
@@ -561,35 +583,45 @@ pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap
 		fz_always(ctx)
 		{
 			fz_drop_pixmap(ctx, original);
-			if (pixmap_cloned)
-				fz_drop_pixmap(ctx, pixmap);
 		}
 		fz_catch(ctx)
+		{
+			if (pixmap_cloned)
+				fz_drop_pixmap(ctx, pixmap);
 			fz_rethrow(ctx);
+		}
 	}
 
-	n = pixmap->n - pixmap->alpha;
-	bpp = pixmap->n;
-	if (fz_colorspace_is_subtractive(ctx, pixmap->colorspace))
-		white = 0;
-	else
-		white = 255;
-
-	inv_ctm = fz_post_scale(fz_invert_matrix(ctm), pixmap->w, pixmap->h);
-	r = fz_round_rect(fz_transform_rect(fz_rect_from_quad(q), inv_ctm));
-	r.x0 = fz_clampi(r.x0, 0, pixmap->w);
-	r.x1 = fz_clampi(r.x1, 0, pixmap->w);
-	r.y1 = fz_clampi(pixmap->h - r.y1, 0, pixmap->h);
-	r.y0 = fz_clampi(pixmap->h - r.y0, 0, pixmap->h);
-	for (y = r.y1; y < r.y0; ++y)
+	/* If we have a 1x1 image, to which a mask is being applied
+	 * then it's the mask we really want to change, not the
+	 * image. We might have just a small section of the image
+	 * being covered, and setting the whole thing to white
+	 * will blank stuff outside the desired area. */
+	if (mask && (pixmap->w > 1 || pixmap->h > 1))
 	{
-		for (x = r.x0; x < r.x1; ++x)
+		n = pixmap->n - pixmap->alpha;
+		bpp = pixmap->n;
+		if (fz_colorspace_is_subtractive(ctx, pixmap->colorspace))
+			white = 0;
+		else
+			white = 255;
+
+		inv_ctm = fz_post_scale(fz_invert_matrix(ctm), pixmap->w, pixmap->h);
+		r = fz_round_rect(fz_transform_rect(fz_rect_from_quad(q), inv_ctm));
+		r.x0 = fz_clampi(r.x0, 0, pixmap->w);
+		r.x1 = fz_clampi(r.x1, 0, pixmap->w);
+		r.y1 = fz_clampi(pixmap->h - r.y1, 0, pixmap->h);
+		r.y0 = fz_clampi(pixmap->h - r.y0, 0, pixmap->h);
+		for (y = r.y1; y < r.y0; ++y)
 		{
-			unsigned char *s = &pixmap->samples[(size_t)y * pixmap->stride + (size_t)x * bpp];
-			for (k = 0; k < n; ++k)
-				s[k] = white;
-			if (pixmap->alpha)
-				s[k] = 255;
+			for (x = r.x0; x < r.x1; ++x)
+			{
+				unsigned char *s = &pixmap->samples[(size_t)y * pixmap->stride + (size_t)x * bpp];
+				for (k = 0; k < n; ++k)
+					s[k] = white;
+				if (pixmap->alpha)
+					s[k] = 255;
+			}
 		}
 	}
 
@@ -859,7 +891,9 @@ pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_o
 	int black_boxes = 0;
 	int image_method = PDF_REDACT_IMAGE_PIXELS;
 
-	pdf_filter_options filter;
+	pdf_filter_options options = { 0 };
+	pdf_sanitize_filter_options sopts = { 0 };
+	pdf_filter_factory list[2] = { 0 };
 
 	if (opts)
 	{
@@ -867,19 +901,23 @@ pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_o
 		image_method = opts->image_method;
 	}
 
-	memset(&filter, 0, sizeof filter);
-	filter.opaque = page;
-	filter.text_filter = pdf_redact_text_filter;
-	if (image_method == PDF_REDACT_IMAGE_PIXELS)
-		filter.image_filter = pdf_redact_image_filter_pixels;
-	if (image_method == PDF_REDACT_IMAGE_REMOVE)
-		filter.image_filter = pdf_redact_image_filter_remove;
+	options.recurse = 0; /* don't redact patterns, softmasks, and type3 fonts */
+	options.instance_forms = 1; /* redact xobjects with instancing */
+	options.ascii = 1;
+	options.opaque = page;
+	options.filters = list;
 	if (black_boxes)
-		filter.end_page = pdf_redact_end_page;
-	filter.recurse = 0; /* don't redact patterns, softmasks, and type3 fonts */
-	filter.instance_forms = 1; /* redact xobjects with instancing */
-	filter.sanitize = 1;
-	filter.ascii = 1;
+		options.complete = pdf_redact_end_page;
+
+	list[0].filter = pdf_new_sanitize_filter;
+	list[0].options = &sopts;
+
+	sopts.opaque = page;
+	sopts.text_filter = pdf_redact_text_filter;
+	if (image_method == PDF_REDACT_IMAGE_PIXELS)
+		sopts.image_filter = pdf_redact_image_filter_pixels;
+	if (image_method == PDF_REDACT_IMAGE_REMOVE)
+		sopts.image_filter = pdf_redact_image_filter_remove;
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
@@ -891,7 +929,7 @@ pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_o
 	pdf_begin_operation(ctx, doc, "Redact page");
 	fz_try(ctx)
 	{
-		pdf_filter_page_contents(ctx, doc, page, &filter);
+		pdf_filter_page_contents(ctx, doc, page, &options);
 		pdf_redact_page_links(ctx, doc, page);
 		pdf_redact_page_annotations(ctx, doc, page);
 
