@@ -1,5 +1,27 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
-#include "mupdf/pdf.h"
+#include "pdf-annot-imp.h"
 
 #include <string.h>
 #include <math.h>
@@ -10,30 +32,44 @@
 void *
 pdf_new_processor(fz_context *ctx, int size)
 {
-	return Memento_label(fz_calloc(ctx, 1, size), "pdf_processor");
+	pdf_processor *ret = Memento_label(fz_calloc(ctx, 1, size), "pdf_processor");
+	ret->refs = 1;
+	return ret;
+}
+
+pdf_processor *
+pdf_keep_processor(fz_context *ctx, pdf_processor *proc)
+{
+	return fz_keep_imp(ctx, proc, &proc->refs);
 }
 
 void
 pdf_close_processor(fz_context *ctx, pdf_processor *proc)
 {
-	if (proc && proc->close_processor)
-	{
-		proc->close_processor(ctx, proc);
-		proc->close_processor = NULL;
-	}
+	void (*close_processor)(fz_context *ctx, pdf_processor *proc);
+
+	if (!proc)
+		return;
+
+	close_processor = proc->close_processor;
+	if (!close_processor)
+		return;
+
+	proc->close_processor = NULL;
+	close_processor(ctx, proc); /* Tail recursion */
 }
 
 void
 pdf_drop_processor(fz_context *ctx, pdf_processor *proc)
 {
-	if (proc)
+	if (fz_drop_imp(ctx, proc, &proc->refs))
 	{
 		if (proc->close_processor)
 			fz_warn(ctx, "dropping unclosed PDF processor");
 		if (proc->drop_processor)
 			proc->drop_processor(ctx, proc);
+		fz_free(ctx, proc);
 	}
-	fz_free(ctx, proc);
 }
 
 static void
@@ -297,11 +333,11 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 			if (tr && !pdf_name_eq(ctx, tr, PDF_NAME(Identity)))
 				fz_warn(ctx, "ignoring transfer function");
 
-			proc->op_gs_SMask(ctx, proc, xobj, csi->rdb, softmask_bc, luminosity);
+			proc->op_gs_SMask(ctx, proc, xobj, softmask_bc, luminosity);
 		}
 		else if (pdf_is_name(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(None)))
 		{
-			proc->op_gs_SMask(ctx, proc, NULL, NULL, NULL, 0);
+			proc->op_gs_SMask(ctx, proc, NULL, NULL, 0);
 		}
 	}
 }
@@ -325,13 +361,13 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 	if (!pdf_is_name(ctx, subtype))
 		fz_throw(ctx, FZ_ERROR_MINOR, "no XObject subtype specified");
 
-	if (pdf_is_hidden_ocg(ctx, csi->doc->ocg, csi->rdb, proc->usage, pdf_dict_get(ctx, xobj, PDF_NAME(OC))))
+	if (pdf_is_ocg_hidden(ctx, csi->doc, csi->rdb, proc->usage, pdf_dict_get(ctx, xobj, PDF_NAME(OC))))
 		return;
 
 	if (pdf_name_eq(ctx, subtype, PDF_NAME(Form)))
 	{
 		if (proc->op_Do_form)
-			proc->op_Do_form(ctx, proc, csi->name, xobj, csi->rdb);
+			proc->op_Do_form(ctx, proc, csi->name, xobj);
 	}
 
 	else if (pdf_name_eq(ctx, subtype, PDF_NAME(Image)))
@@ -357,6 +393,8 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 static void
 pdf_process_CS(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 {
+	fz_colorspace *cs;
+
 	if (!proc->op_CS || !proc->op_cs)
 		return;
 
@@ -366,39 +404,44 @@ pdf_process_CS(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 			proc->op_CS(ctx, proc, "Pattern", NULL);
 		else
 			proc->op_cs(ctx, proc, "Pattern", NULL);
+		return;
 	}
+
+	if (!strcmp(csi->name, "DeviceGray"))
+		cs = fz_keep_colorspace(ctx, fz_device_gray(ctx));
+	else if (!strcmp(csi->name, "DeviceRGB"))
+		cs = fz_keep_colorspace(ctx, fz_device_rgb(ctx));
+	else if (!strcmp(csi->name, "DeviceCMYK"))
+		cs = fz_keep_colorspace(ctx, fz_device_cmyk(ctx));
 	else
 	{
-		fz_colorspace *cs;
-
-		if (!strcmp(csi->name, "DeviceGray"))
-			cs = fz_keep_colorspace(ctx, fz_device_gray(ctx));
-		else if (!strcmp(csi->name, "DeviceRGB"))
-			cs = fz_keep_colorspace(ctx, fz_device_rgb(ctx));
-		else if (!strcmp(csi->name, "DeviceCMYK"))
-			cs = fz_keep_colorspace(ctx, fz_device_cmyk(ctx));
-		else
-		{
-			pdf_obj *csres, *csobj;
-			csres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ColorSpace));
-			csobj = pdf_dict_gets(ctx, csres, csi->name);
-			if (!csobj)
-				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ColorSpace resource '%s'", csi->name);
-			cs = pdf_load_colorspace(ctx, csobj);
-		}
-
-		fz_try(ctx)
+		pdf_obj *csres, *csobj;
+		csres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ColorSpace));
+		csobj = pdf_dict_gets(ctx, csres, csi->name);
+		if (!csobj)
+			fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ColorSpace resource '%s'", csi->name);
+		if (pdf_is_array(ctx, csobj) && pdf_array_len(ctx, csobj) == 1 && pdf_name_eq(ctx, pdf_array_get(ctx, csobj, 0), PDF_NAME(Pattern)))
 		{
 			if (stroke)
-				proc->op_CS(ctx, proc, csi->name, cs);
+				proc->op_CS(ctx, proc, "Pattern", NULL);
 			else
-				proc->op_cs(ctx, proc, csi->name, cs);
+				proc->op_cs(ctx, proc, "Pattern", NULL);
+			return;
 		}
-		fz_always(ctx)
-			fz_drop_colorspace(ctx, cs);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
+		cs = pdf_load_colorspace(ctx, csobj);
 	}
+
+	fz_try(ctx)
+	{
+		if (stroke)
+			proc->op_CS(ctx, proc, csi->name, cs);
+		else
+			proc->op_cs(ctx, proc, csi->name, cs);
+	}
+	fz_always(ctx)
+		fz_drop_colorspace(ctx, cs);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
@@ -497,7 +540,7 @@ pdf_process_BDC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 	if (strcmp(csi->name, "OC"))
 		return;
 
-	if (pdf_is_hidden_ocg(ctx, csi->doc->ocg, csi->rdb, proc->usage, csi->obj))
+	if (pdf_is_ocg_hidden(ctx, csi->doc, csi->rdb, proc->usage, csi->obj))
 		++proc->hidden;
 }
 
@@ -1020,8 +1063,18 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 	while (tok != PDF_TOK_EOF);
 }
 
+void pdf_processor_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *res)
+{
+	proc->push_resources(ctx, proc, res);
+}
+
+pdf_obj *pdf_processor_pop_resources(fz_context *ctx, pdf_processor *proc)
+{
+	return proc->pop_resources(ctx, proc);
+}
+
 void
-pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *rdb, pdf_obj *stmobj, fz_cookie *cookie)
+pdf_process_raw_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *rdb, pdf_obj *stmobj, fz_cookie *cookie)
 {
 	pdf_csi csi;
 	pdf_lexbuf buf;
@@ -1056,6 +1109,24 @@ pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pd
 	}
 }
 
+void
+pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_obj *rdb, pdf_obj *stmobj, fz_cookie *cookie, pdf_obj **out_res)
+{
+	pdf_processor_push_resources(ctx, proc, rdb);
+	fz_try(ctx)
+		pdf_process_raw_contents(ctx, proc, doc, rdb, stmobj, cookie);
+	fz_always(ctx)
+	{
+		pdf_obj *res = pdf_processor_pop_resources(ctx, proc);
+		if (out_res)
+			*out_res = res;
+		else
+			pdf_drop_obj(ctx, res);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
 /* Bug 702543: It looks like certain types of annotation are never
  * printed. */
 static int
@@ -1071,11 +1142,11 @@ pdf_should_print_annot(fz_context *ctx, pdf_annot *annot)
 }
 
 void
-pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_page *page, pdf_annot *annot, fz_cookie *cookie)
+pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_annot *annot, fz_cookie *cookie)
 {
 	int flags = pdf_dict_get_int(ctx, annot->obj, PDF_NAME(F));
 
-	if (flags & (PDF_ANNOT_IS_INVISIBLE | PDF_ANNOT_IS_HIDDEN))
+	if (flags & (PDF_ANNOT_IS_INVISIBLE | PDF_ANNOT_IS_HIDDEN) || annot->hidden_editing)
 		return;
 
 	/* popup annotations should never be drawn */
@@ -1098,18 +1169,24 @@ pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_p
 	/* TODO: NoZoom and NoRotate */
 
 	/* XXX what resources, if any, to use for this check? */
-	if (pdf_is_hidden_ocg(ctx, doc->ocg, NULL, proc->usage, pdf_dict_get(ctx, annot->obj, PDF_NAME(OC))))
+	if (pdf_is_ocg_hidden(ctx, annot->page->doc, NULL, proc->usage, pdf_dict_get(ctx, annot->obj, PDF_NAME(OC))))
 		return;
 
-	if (proc->op_q && proc->op_cm && proc->op_Do_form && proc->op_Q && annot->ap)
+	if (proc->op_q && proc->op_cm && proc->op_Do_form && proc->op_Q)
 	{
-		fz_matrix matrix = pdf_annot_transform(ctx, annot);
+		pdf_obj *ap = pdf_annot_ap(ctx, annot);
+		fz_matrix matrix;
+
+		if (!ap)
+			return;
+
+		matrix = pdf_annot_transform(ctx, annot);
 		proc->op_q(ctx, proc);
 		proc->op_cm(ctx, proc,
 			matrix.a, matrix.b,
 			matrix.c, matrix.d,
 			matrix.e, matrix.f);
-		proc->op_Do_form(ctx, proc, NULL, annot->ap, pdf_page_resources(ctx, page));
+		proc->op_Do_form(ctx, proc, NULL, ap);
 		proc->op_Q(ctx, proc);
 	}
 }
@@ -1131,12 +1208,14 @@ pdf_process_glyph(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_o
 
 	fz_try(ctx)
 	{
+		pdf_processor_push_resources(ctx, proc, rdb);
 		stm = fz_open_buffer(ctx, contents);
 		pdf_process_stream(ctx, proc, &csi, stm);
 		pdf_process_end(ctx, proc, &csi);
 	}
 	fz_always(ctx)
 	{
+		pdf_drop_obj(ctx, pdf_processor_pop_resources(ctx, proc));
 		fz_drop_stream(ctx, stm);
 		pdf_clear_stack(ctx, &csi);
 		pdf_lexbuf_fin(ctx, &buf);
