@@ -468,6 +468,9 @@ int pdf_to_gen(fz_context *ctx, pdf_obj *obj)
 	return 0;
 }
 
+/*
+	DEPRECATED: Do not use in new code.
+*/
 pdf_document *pdf_get_indirect_document(fz_context *ctx, pdf_obj *obj)
 {
 	if (OBJ_IS_INDIRECT(obj))
@@ -475,6 +478,9 @@ pdf_document *pdf_get_indirect_document(fz_context *ctx, pdf_obj *obj)
 	return NULL;
 }
 
+/*
+	DEPRECATED: Do not use in new code.
+*/
 pdf_document *pdf_get_bound_document(fz_context *ctx, pdf_obj *obj)
 {
 	if (obj < PDF_LIMIT)
@@ -488,6 +494,16 @@ pdf_document *pdf_get_bound_document(fz_context *ctx, pdf_obj *obj)
 	return NULL;
 }
 
+/*
+	This implementation will do to provide the required
+	API change in advance of the rewrite to use weak references
+	in the next version.
+*/
+pdf_document *pdf_pin_document(fz_context *ctx, pdf_obj *obj)
+{
+	return pdf_keep_document(ctx, pdf_get_bound_document(ctx, obj));
+}
+
 int pdf_objcmp_resolve(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 {
 	RESOLVE(a);
@@ -495,10 +511,10 @@ int pdf_objcmp_resolve(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 	return pdf_objcmp(ctx, a, b);
 }
 
-int
-pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+static int
+do_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b, int check_streams)
 {
-	int i;
+	int i, j;
 
 	if (a == b)
 		return 0;
@@ -575,33 +591,98 @@ pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 	case PDF_DICT:
 		if (DICT(a)->len != DICT(b)->len)
 			return DICT(a)->len - DICT(b)->len;
-		for (i = 0; i < DICT(a)->len; i++)
+		if ((a->flags & b->flags) & PDF_FLAGS_SORTED)
 		{
-			if (pdf_objcmp(ctx, DICT(a)->items[i].k, DICT(b)->items[i].k))
-				return 1;
-			if (pdf_objcmp(ctx, DICT(a)->items[i].v, DICT(b)->items[i].v))
-				return 1;
+			/* Both a and b are sorted. Easy. */
+			for (i = 0; i < DICT(a)->len; i++)
+			{
+				if (pdf_objcmp(ctx, DICT(a)->items[i].k, DICT(b)->items[i].k))
+					return 1;
+				if (pdf_objcmp(ctx, DICT(a)->items[i].v, DICT(b)->items[i].v))
+					return 1;
+			}
+		}
+		else
+		{
+			/* Either a or b is not sorted. We need to work harder. */
+			int len = DICT(a)->len;
+			for (i = 0; i < len; i++)
+			{
+				pdf_obj *key = DICT(a)->items[i].k;
+				pdf_obj *val = DICT(a)->items[i].v;
+				for (j = 0; j < len; j++)
+				{
+					if (pdf_objcmp(ctx, key, DICT(b)->items[j].k) == 0 &&
+						pdf_objcmp(ctx, val, DICT(b)->items[j].v) == 0)
+						break; /* Match */
+				}
+				if (j == len)
+					return 1;
+			}
 		}
 		/* Dicts are identical, but if they are streams, we can only be sure
-		 * they are identical if the stream contents match. We don't currently
-		 * test for identical stream contents, so if they are streams, require
-		 * a to == b for a match. */
+		 * they are identical if the stream contents match. If '!check_streams',
+		 * then don't test for identical stream contents - only match if a == b.
+		 * Otherwise, do the full, painful, comparison. */
 		{
 			/* Slightly convoluted to know if something is a stream. */
 			pdf_document *doc = DICT(a)->doc;
-			int n = pdf_obj_parent_num(ctx, a);
-			pdf_xref_entry *entry = pdf_get_xref_entry(ctx, doc, n);
-			if (entry->obj == a && pdf_obj_num_is_stream(ctx, doc, n))
+			int ap = pdf_obj_parent_num(ctx, a);
+			int bp;
+			int a_is_stream = 0;
+			pdf_xref_entry *entrya = pdf_get_xref_entry_no_change(ctx, doc, ap);
+			pdf_xref_entry *entryb;
+			if (entrya != NULL && entrya->obj == a && pdf_obj_num_is_stream(ctx, doc, ap))
 			{
-				/* It's a stream, and we know a != b from above. So mismatch. */
-				return 1;
+				/* It's a stream, and we know a != b from above. */
+				if (!check_streams)
+					return 1; /* mismatch */
+				a_is_stream = 1;
 			}
-			n = pdf_obj_parent_num(ctx, b);
-			entry = pdf_get_xref_entry(ctx, doc, n);
-			if (entry->obj == b && pdf_obj_num_is_stream(ctx, doc, n))
+			bp = pdf_obj_parent_num(ctx, b);
+			entryb = pdf_get_xref_entry_no_change(ctx, doc, bp);
+			if (entryb != NULL && entryb->obj == b && pdf_obj_num_is_stream(ctx, doc, bp))
 			{
 				/* It's a stream, and we know a != b from above. So mismatch. */
-				return 1;
+				if (!check_streams || !a_is_stream)
+					return 1; /* mismatch */
+			}
+			else
+			{
+				/* b is not a stream. We match, iff a is not a stream. */
+				return a_is_stream;
+			}
+			/* So, if we get here, we know check_streams is true, and that both
+			 * a and b are streams. */
+			{
+				fz_buffer *sa = NULL;
+				fz_buffer *sb = NULL;
+				int differ = 1;
+
+				fz_var(sa);
+				fz_var(sb);
+
+				fz_try(ctx)
+				{
+					unsigned char *dataa, *datab;
+					size_t lena, lenb;
+					sa = pdf_load_raw_stream_number(ctx, doc, ap);
+					sb = pdf_load_raw_stream_number(ctx, doc, bp);
+					lena = fz_buffer_storage(ctx, sa, &dataa);
+					lenb = fz_buffer_storage(ctx, sb, &datab);
+					if (lena == lenb && memcmp(dataa, datab, lena) == 0)
+						differ = 0;
+				}
+				fz_always(ctx)
+				{
+					fz_drop_buffer(ctx, sa);
+					fz_drop_buffer(ctx, sb);
+				}
+				fz_catch(ctx)
+				{
+					fz_rethrow(ctx);
+				}
+				return differ;
 			}
 		}
 
@@ -610,6 +691,18 @@ pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 	return 1;
 }
 
+int
+pdf_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+{
+	return do_objcmp(ctx, a, b, 0);
+}
+
+int
+pdf_objcmp_deep(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+{
+	return do_objcmp(ctx, a, b, 1);
+}
+
 int pdf_name_eq(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 {
 	RESOLVE(a);
@@ -1405,6 +1498,27 @@ static void prepare_object_for_alteration(fz_context *ctx, pdf_obj *obj, pdf_obj
 		return;
 	}
 
+	/* Do we need to drop the page maps? */
+	if (doc && (doc->rev_page_map || doc->fwd_page_map))
+	{
+		if (doc->non_structural_change)
+		{
+			/* No need to drop the reverse page map on a non-structural change. */
+		}
+		else if (parent == 0)
+		{
+			/* This object isn't linked into the document - can't change the
+			 * pagemap. */
+		}
+		else if (doc->local_xref && doc->local_xref_nesting > 0)
+		{
+			/* We have a local_xref and it's in force. By convention, we
+			 * never do structural changes in local_xrefs. */
+		}
+		else
+			pdf_drop_page_tree_internal(ctx, doc);
+	}
+
 	if (val)
 	{
 		val_doc = pdf_get_bound_document(ctx, val);
@@ -2017,10 +2131,11 @@ pdf_obj *
 pdf_dict_geta(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *abbrev)
 {
 	pdf_obj *v;
-	v = pdf_dict_get(ctx, obj, key);
+	/* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. */
+	v = pdf_dict_get(ctx, obj, abbrev);
 	if (v)
 		return v;
-	return pdf_dict_get(ctx, obj, abbrev);
+	return pdf_dict_get(ctx, obj, key);
 }
 
 static void
@@ -2502,6 +2617,8 @@ pdf_mark_list_push(fz_context *ctx, pdf_mark_list *marks, pdf_obj *obj)
 	int num = pdf_to_num(ctx, obj);
 	int i;
 
+	/* If object is not an indirection, then no record to check.
+	 * We must still push it to allow pops to stay in sync. */
 	if (num > 0)
 	{
 		/* Note: this is slow, if the mark list is expected to be big use pdf_mark_bits instead! */
@@ -2514,7 +2631,10 @@ pdf_mark_list_push(fz_context *ctx, pdf_mark_list *marks, pdf_obj *obj)
 	{
 		int newsize = marks->max << 1;
 		if (marks->list == marks->local_list)
+		{
 			marks->list = fz_malloc_array(ctx, newsize, int);
+			memcpy(marks->list, marks->local_list, sizeof(marks->local_list));
+		}
 		else
 			marks->list = fz_realloc_array(ctx, marks->list, newsize, int);
 		marks->max = newsize;
@@ -3169,50 +3289,58 @@ int pdf_obj_refs(fz_context *ctx, pdf_obj *obj)
 
 /* Convenience functions */
 
-static pdf_obj *
-pdf_dict_get_inheritable_imp(fz_context *ctx, pdf_obj *node, pdf_obj *key, int depth, pdf_cycle_list *cycle_up)
-{
-	pdf_cycle_list cycle;
-	pdf_obj *val = pdf_dict_get(ctx, node, key);
-	if (val)
-		return val;
-	if (pdf_cycle(ctx, &cycle, cycle_up, node))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
-	if (depth > 100)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "too much recursion in tree (parents)");
-	node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
-	if (node)
-		return pdf_dict_get_inheritable_imp(ctx, node, key, depth + 1, &cycle);
-	return NULL;
-}
+/*
+	Uses Floyd's cycle finding algorithm, modified to avoid starting
+	the 'slow' pointer for a while.
 
+	https://www.geeksforgeeks.org/floyds-cycle-finding-algorithm/
+*/
 pdf_obj *
 pdf_dict_get_inheritable(fz_context *ctx, pdf_obj *node, pdf_obj *key)
 {
-	return pdf_dict_get_inheritable_imp(ctx, node, key, 0, NULL);
-}
+	pdf_obj *slow = node;
+	int halfbeat = 11; /* Don't start moving slow pointer for a while. */
+
+	while (node)
+	{
+		pdf_obj *val = pdf_dict_get(ctx, node, key);
+		if (val)
+			return val;
+		node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
+		if (node == slow)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in resources");
+		if (--halfbeat == 0)
+		{
+			slow = pdf_dict_get(ctx, slow, PDF_NAME(Parent));
+			halfbeat = 2;
+		}
+	}
 
-static pdf_obj *
-pdf_dict_getp_inheritable_imp(fz_context *ctx, pdf_obj *node, const char *path, int depth, pdf_cycle_list *cycle_up)
-{
-	pdf_cycle_list cycle;
-	pdf_obj *val = pdf_dict_getp(ctx, node, path);
-	if (val)
-		return val;
-	if (pdf_cycle(ctx, &cycle, cycle_up, node))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in tree (parents)");
-	if (depth > 100)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "too much recursion in tree (parents)");
-	node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
-	if (node)
-		return pdf_dict_getp_inheritable_imp(ctx, node, path, depth + 1, &cycle);
 	return NULL;
 }
 
 pdf_obj *
 pdf_dict_getp_inheritable(fz_context *ctx, pdf_obj *node, const char *path)
 {
-	return pdf_dict_getp_inheritable_imp(ctx, node, path, 0, NULL);
+	pdf_obj *slow = node;
+	int halfbeat = 11; /* Don't start moving slow pointer for a while. */
+
+	while (node)
+	{
+		pdf_obj *val = pdf_dict_getp(ctx, node, path);
+		if (val)
+			return val;
+		node = pdf_dict_get(ctx, node, PDF_NAME(Parent));
+		if (node == slow)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in resources");
+		if (--halfbeat == 0)
+		{
+			slow = pdf_dict_get(ctx, slow, PDF_NAME(Parent));
+			halfbeat = 2;
+		}
+	}
+
+	return NULL;
 }
 
 void pdf_dict_put_bool(fz_context *ctx, pdf_obj *dict, pdf_obj *key, int x)
@@ -3409,3 +3537,15 @@ fz_matrix pdf_array_get_matrix(fz_context *ctx, pdf_obj *array, int index)
 {
 	return pdf_to_matrix(ctx, pdf_array_get(ctx, array, index));
 }
+
+#ifndef NDEBUG
+void pdf_verify_name_table_sanity(void)
+{
+	int i;
+
+	for (i = PDF_ENUM_FALSE+1; i < PDF_ENUM_LIMIT-1; i++)
+	{
+		assert(strcmp(PDF_NAME_LIST[i], PDF_NAME_LIST[i+1]) < 0);
+	}
+}
+#endif
