@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -24,6 +24,7 @@
 #include "mupdf/pdf.h"
 
 #include <string.h>
+#include <math.h>
 
 /*
 	The URI encoding format broadly follows that described in
@@ -70,24 +71,32 @@ pdf_test_outline(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_mark_bit
 			if (*fixed == 0)
 				pdf_begin_operation(ctx, doc, "Repair outline nodes");
 			*fixed = 1;
-		}
-		if (parent_diff)
-		{
-			fz_warn(ctx, "Bad or missing parent pointer in outline tree, repairing");
-			pdf_dict_put(ctx, dict, PDF_NAME(Parent), expected_parent);
-		}
-		if (prev_diff)
-		{
-			fz_warn(ctx, "Bad or missing prev pointer in outline tree, repairing");
-			if (expected_prev)
-				pdf_dict_put(ctx, dict, PDF_NAME(Prev), expected_prev);
-			else
-				pdf_dict_del(ctx, dict, PDF_NAME(Prev));
-		}
-		if (last_diff)
-		{
-			fz_warn(ctx, "Bad or missing last pointer in outline tree, repairing");
-			pdf_dict_put(ctx, expected_parent, PDF_NAME(Last), dict);
+			doc->non_structural_change = 1;
+			fz_try(ctx)
+			{
+				if (parent_diff)
+				{
+					fz_warn(ctx, "Bad or missing parent pointer in outline tree, repairing");
+					pdf_dict_put(ctx, dict, PDF_NAME(Parent), expected_parent);
+				}
+				if (prev_diff)
+				{
+					fz_warn(ctx, "Bad or missing prev pointer in outline tree, repairing");
+					if (expected_prev)
+						pdf_dict_put(ctx, dict, PDF_NAME(Prev), expected_prev);
+					else
+						pdf_dict_del(ctx, dict, PDF_NAME(Prev));
+				}
+				if (last_diff)
+				{
+					fz_warn(ctx, "Bad or missing last pointer in outline tree, repairing");
+					pdf_dict_put(ctx, expected_parent, PDF_NAME(Last), dict);
+				}
+			}
+			fz_always(ctx)
+				doc->non_structural_change = 0;
+			fz_catch(ctx)
+				fz_rethrow(ctx);
 		}
 
 		first = pdf_dict_get(ctx, dict, PDF_NAME(First));
@@ -511,7 +520,8 @@ fz_outline_iterator *pdf_new_outline_iterator(fz_context *ctx, pdf_document *doc
 		first = pdf_dict_get(ctx, obj, PDF_NAME(First));
 		if (first)
 		{
-			/* cache page tree for fast link destination lookups */
+			/* cache page tree for fast link destination lookups. This
+			 * will be dropped 'just in time' on writes to the doc. */
 			pdf_load_page_tree(ctx, doc);
 			fz_try(ctx)
 			{
@@ -530,7 +540,6 @@ fz_outline_iterator *pdf_new_outline_iterator(fz_context *ctx, pdf_document *doc
 			{
 				if (fixed)
 					pdf_end_operation(ctx, doc);
-				pdf_drop_page_tree(ctx, doc);
 			}
 			fz_catch(ctx)
 				fz_rethrow(ctx);
@@ -620,29 +629,50 @@ pdf_new_destination_from_link(fz_context *ctx, pdf_document *doc, const char *ur
 		case FZ_LINK_DEST_FIT_H:
 			p = fz_transform_point_xy(0, val.y, invctm);
 			pdf_array_push(ctx, dest, PDF_NAME(FitH));
-			pdf_array_push_real(ctx, dest, p.y);
+			if (isnan(p.y))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.y);
 			break;
 		case FZ_LINK_DEST_FIT_BH:
 			p = fz_transform_point_xy(0, val.y, invctm);
 			pdf_array_push(ctx, dest, PDF_NAME(FitBH));
-			pdf_array_push_real(ctx, dest, p.y);
+			if (isnan(p.y))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.y);
 			break;
 		case FZ_LINK_DEST_FIT_V:
 			p = fz_transform_point_xy(val.x, 0, invctm);
 			pdf_array_push(ctx, dest, PDF_NAME(FitV));
-			pdf_array_push_real(ctx, dest, p.x);
+			if (isnan(p.x))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.x);
 			break;
 		case FZ_LINK_DEST_FIT_BV:
 			p = fz_transform_point_xy(val.x, 0, invctm);
 			pdf_array_push(ctx, dest, PDF_NAME(FitBV));
-			pdf_array_push_real(ctx, dest, p.x);
+			if (isnan(p.x))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.x);
 			break;
 		case FZ_LINK_DEST_XYZ:
 			p = fz_transform_point_xy(val.x, val.y, invctm);
 			pdf_array_push(ctx, dest, PDF_NAME(XYZ));
-			pdf_array_push_real(ctx, dest, p.x);
-			pdf_array_push_real(ctx, dest, p.y);
-			pdf_array_push_real(ctx, dest, val.zoom / 100);
+			if (isnan(p.x))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.x);
+			if (isnan(p.y))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, p.y);
+			if (isnan(val.zoom))
+				pdf_array_push(ctx, dest, PDF_NULL);
+			else
+				pdf_array_push_real(ctx, dest, val.zoom / 100);
 			break;
 		case FZ_LINK_DEST_FIT_R:
 			r.x0 = val.x;
