@@ -1,5 +1,27 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
-#include "mupdf/pdf.h"
+#include "pdf-annot-imp.h"
 
 static void
 pdf_run_annot_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_annot *annot, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
@@ -9,9 +31,32 @@ pdf_run_annot_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf
 	pdf_processor *proc = NULL;
 	fz_default_colorspaces *default_cs = NULL;
 	int flags;
+	int resources_pushed = 0;
+	int struct_parent_num = -1;
+	pdf_obj *struct_parent;
 
 	fz_var(proc);
 	fz_var(default_cs);
+	fz_var(resources_pushed);
+
+	if (cookie && page->super.incomplete)
+		cookie->incomplete = 1;
+
+	pdf_annot_push_local_xref(ctx, annot);
+
+	/* Widgets only get displayed if they have both a T and a TF flag,
+	 * apparently */
+	if (pdf_name_eq(ctx, pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)), PDF_NAME(Widget)))
+	{
+		pdf_obj *ft = pdf_dict_get_inheritable(ctx, annot->obj, PDF_NAME(FT));
+		pdf_obj *t = pdf_dict_get_inheritable(ctx, annot->obj, PDF_NAME(T));
+
+		if (ft == NULL || t == NULL)
+		{
+			pdf_annot_pop_local_xref(ctx, annot);
+			return;
+		}
+	}
 
 	fz_try(ctx)
 	{
@@ -34,21 +79,30 @@ pdf_run_annot_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf
 
 		ctm = fz_concat(page_ctm, ctm);
 
-		proc = pdf_new_run_processor(ctx, dev, ctm, usage, NULL, default_cs);
-		pdf_process_annot(ctx, proc, doc, page, annot, cookie);
+		struct_parent = pdf_dict_getl(ctx, page->obj, PDF_NAME(StructParent));
+		if (pdf_is_number(ctx, struct_parent))
+			struct_parent_num = pdf_to_int(ctx, struct_parent);
+
+		proc = pdf_new_run_processor(ctx, page->doc, dev, ctm, struct_parent_num, usage, NULL, default_cs, cookie);
+		pdf_processor_push_resources(ctx, proc, pdf_page_resources(ctx, annot->page));
+		resources_pushed = 1;
+		pdf_process_annot(ctx, proc, annot, cookie);
 		pdf_close_processor(ctx, proc);
 	}
 	fz_always(ctx)
 	{
+		if (resources_pushed)
+			pdf_processor_pop_resources(ctx, proc);
 		pdf_drop_processor(ctx, proc);
 		fz_drop_default_colorspaces(ctx, default_cs);
+		pdf_annot_pop_local_xref(ctx, annot);
 	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
 
 static void
-pdf_run_page_contents_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
+pdf_run_page_contents_with_usage_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
 	fz_matrix page_ctm;
 	pdf_obj *resources;
@@ -57,11 +111,16 @@ pdf_run_page_contents_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *p
 	pdf_processor *proc = NULL;
 	fz_default_colorspaces *default_cs = NULL;
 	fz_colorspace *colorspace = NULL;
+	int struct_parent_num = -1;
+	pdf_obj *struct_parent;
 
 	fz_var(proc);
 	fz_var(colorspace);
 	fz_var(default_cs);
 
+	if (cookie && page->super.incomplete)
+		cookie->incomplete = 1;
+
 	fz_try(ctx)
 	{
 		default_cs = pdf_load_default_colorspaces(ctx, doc, page);
@@ -87,7 +146,16 @@ pdf_run_page_contents_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *p
 					fz_try(ctx)
 						colorspace = pdf_load_colorspace(ctx, cs);
 					fz_catch(ctx)
+					{
+						fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+						fz_warn(ctx, "Ignoring Page blending colorspace.");
+					}
+					if (!fz_is_valid_blend_colorspace(ctx, colorspace))
+					{
+						fz_warn(ctx, "Ignoring invalid Page blending colorspace: %s.", colorspace->name);
+						fz_drop_colorspace(ctx, colorspace);
 						colorspace = NULL;
+					}
 				}
 			}
 			else
@@ -96,8 +164,12 @@ pdf_run_page_contents_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *p
 			fz_begin_group(ctx, dev, mediabox, colorspace, 1, 0, 0, 1);
 		}
 
-		proc = pdf_new_run_processor(ctx, dev, ctm, usage, NULL, default_cs);
-		pdf_process_contents(ctx, proc, doc, resources, contents, cookie);
+		struct_parent = pdf_dict_get(ctx, page->obj, PDF_NAME(StructParents));
+		if (pdf_is_number(ctx, struct_parent))
+			struct_parent_num = pdf_to_int(ctx, struct_parent);
+
+		proc = pdf_new_run_processor(ctx, page->doc, dev, ctm, struct_parent_num, usage, NULL, default_cs, cookie);
+		pdf_process_contents(ctx, proc, doc, resources, contents, cookie, NULL);
 		pdf_close_processor(ctx, proc);
 
 		if (page->transparency)
@@ -117,18 +189,7 @@ pdf_run_page_contents_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *p
 	}
 }
 
-/*
-	Interpret a loaded page and render it on a device.
-	Just the main page contents without the annotations
-
-	page: A page loaded by pdf_load_page.
-
-	dev: Device used for rendering, obtained from fz_new_*_device.
-
-	ctm: A transformation matrix applied to the objects on the page,
-	e.g. to scale or rotate the page contents as desired.
-*/
-void pdf_run_page_contents(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
+void pdf_run_page_contents_with_usage(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
 	pdf_document *doc = page->doc;
 	int nocache;
@@ -139,7 +200,7 @@ void pdf_run_page_contents(fz_context *ctx, pdf_page *page, fz_device *dev, fz_m
 
 	fz_try(ctx)
 	{
-		pdf_run_page_contents_with_usage(ctx, doc, page, dev, ctm, "View", cookie);
+		pdf_run_page_contents_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
 	}
 	fz_always(ctx)
 	{
@@ -152,18 +213,11 @@ void pdf_run_page_contents(fz_context *ctx, pdf_page *page, fz_device *dev, fz_m
 	}
 }
 
-/*
-	Interpret an annotation and render it on a device.
+void pdf_run_page_contents(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
+{
+	pdf_run_page_contents_with_usage(ctx, page, dev, ctm, "View", cookie);
+}
 
-	page: A page loaded by pdf_load_page.
-
-	annot: an annotation.
-
-	dev: Device used for rendering, obtained from fz_new_*_device.
-
-	ctm: A transformation matrix applied to the objects on the page,
-	e.g. to scale or rotate the page contents as desired.
-*/
 void pdf_run_annot(fz_context *ctx, pdf_annot *annot, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
 {
 	pdf_page *page = annot->page;
@@ -189,11 +243,11 @@ void pdf_run_annot(fz_context *ctx, pdf_annot *annot, fz_device *dev, fz_matrix
 }
 
 static void
-pdf_run_page_widgets_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
+pdf_run_page_widgets_with_usage_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
-	pdf_widget *widget;
+	pdf_annot *widget;
 
-	if (cookie && cookie->progress_max != -1)
+	if (cookie && cookie->progress_max != (size_t)-1)
 	{
 		int count = 1;
 		for (widget = page->widgets; widget; widget = widget->next)
@@ -216,11 +270,11 @@ pdf_run_page_widgets_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *pa
 }
 
 static void
-pdf_run_page_annots_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
+pdf_run_page_annots_with_usage_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
 	pdf_annot *annot;
 
-	if (cookie && cookie->progress_max != -1)
+	if (cookie && cookie->progress_max != (size_t)-1)
 	{
 		int count = 1;
 		for (annot = page->annots; annot; annot = annot->next)
@@ -242,7 +296,7 @@ pdf_run_page_annots_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *pag
 	}
 }
 
-void pdf_run_page_annots(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
+void pdf_run_page_annots_with_usage(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
 	pdf_document *doc = page->doc;
 	int nocache;
@@ -253,7 +307,36 @@ void pdf_run_page_annots(fz_context *ctx, pdf_page *page, fz_device *dev, fz_mat
 
 	fz_try(ctx)
 	{
-		pdf_run_page_annots_with_usage(ctx, doc, page, dev, ctm, "View", cookie);
+		pdf_run_page_annots_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
+	}
+	fz_always(ctx)
+	{
+		if (nocache)
+			pdf_clear_xref_to_mark(ctx, doc);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+}
+
+void pdf_run_page_annots(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
+{
+	pdf_run_page_annots_with_usage(ctx, page, dev, ctm, "View", cookie);
+}
+
+void pdf_run_page_widgets_with_usage(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
+{
+	pdf_document *doc = page->doc;
+	int nocache;
+
+	nocache = !!(dev->hints & FZ_NO_CACHE);
+	if (nocache)
+		pdf_mark_xref(ctx, doc);
+
+	fz_try(ctx)
+	{
+		pdf_run_page_widgets_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
 	}
 	fz_always(ctx)
 	{
@@ -268,55 +351,22 @@ void pdf_run_page_annots(fz_context *ctx, pdf_page *page, fz_device *dev, fz_mat
 
 void pdf_run_page_widgets(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
 {
-	pdf_document *doc = page->doc;
-	int nocache;
-
-	nocache = !!(dev->hints & FZ_NO_CACHE);
-	if (nocache)
-		pdf_mark_xref(ctx, doc);
-
-	fz_try(ctx)
-	{
-		pdf_run_page_widgets_with_usage(ctx, doc, page, dev, ctm, "View", cookie);
-	}
-	fz_always(ctx)
-	{
-		if (nocache)
-			pdf_clear_xref_to_mark(ctx, doc);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
+	pdf_run_page_widgets_with_usage(ctx, page, dev, ctm, "View", cookie);
 }
 
-/*
-	Interpret a loaded page and render it on a device.
-
-	page: A page loaded by pdf_load_page.
-
-	dev: Device used for rendering, obtained from fz_new_*_device.
-
-	ctm: A transformation matrix applied to the objects on the page,
-	e.g. to scale or rotate the page contents as desired.
-
-	usage: The 'usage' for displaying the file (typically
-	'View', 'Print' or 'Export'). NULL means 'View'.
-
-	cookie: A pointer to an optional fz_cookie structure that can be used
-	to track progress, collect errors etc.
-*/
 void
-pdf_run_page_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
+pdf_run_page_with_usage(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, const char *usage, fz_cookie *cookie)
 {
+	pdf_document *doc = page->doc;
 	int nocache = !!(dev->hints & FZ_NO_CACHE);
 
 	if (nocache)
 		pdf_mark_xref(ctx, doc);
 	fz_try(ctx)
 	{
-		pdf_run_page_contents_with_usage(ctx, doc, page, dev, ctm, usage, cookie);
-		pdf_run_page_annots_with_usage(ctx, doc, page, dev, ctm, usage, cookie);
+		pdf_run_page_contents_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
+		pdf_run_page_annots_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
+		pdf_run_page_widgets_with_usage_imp(ctx, doc, page, dev, ctm, usage, cookie);
 	}
 	fz_always(ctx)
 	{
@@ -329,21 +379,10 @@ pdf_run_page_with_usage(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_d
 	}
 }
 
-/*
-	Interpret a loaded page and render it on a device.
-
-	page: A page loaded by pdf_load_page.
-
-	dev: Device used for rendering, obtained from fz_new_*_device.
-
-	ctm: A transformation matrix applied to the objects on the page,
-	e.g. to scale or rotate the page contents as desired.
-*/
 void
 pdf_run_page(fz_context *ctx, pdf_page *page, fz_device *dev, fz_matrix ctm, fz_cookie *cookie)
 {
-	pdf_document *doc = page->doc;
-	pdf_run_page_with_usage(ctx, doc, page, dev, ctm, "View", cookie);
+	pdf_run_page_with_usage(ctx, page, dev, ctm, "View", cookie);
 }
 
 void
@@ -351,7 +390,7 @@ pdf_run_glyph(fz_context *ctx, pdf_document *doc, pdf_obj *resources, fz_buffer
 {
 	pdf_processor *proc;
 
-	proc = pdf_new_run_processor(ctx, dev, ctm, "View", gstate, default_cs);
+	proc = pdf_new_run_processor(ctx, doc, dev, ctm, -1, "View", gstate, default_cs, NULL);
 	fz_try(ctx)
 	{
 		pdf_process_glyph(ctx, proc, doc, resources, contents);
