@@ -1,11 +1,30 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
 #include <string.h>
 
-/*
- * Check if an object is a stream or not.
- */
 int
 pdf_obj_num_is_stream(fz_context *ctx, pdf_document *doc, int num)
 {
@@ -14,7 +33,13 @@ pdf_obj_num_is_stream(fz_context *ctx, pdf_document *doc, int num)
 	if (num <= 0 || num >= pdf_xref_len(ctx, doc))
 		return 0;
 
-	entry = pdf_cache_object(ctx, doc, num);
+	fz_try(ctx)
+		entry = pdf_cache_object(ctx, doc, num);
+	fz_catch(ctx)
+	{
+		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		return 0;
+	}
 
 	return entry->stm_ofs != 0 || entry->stm_buf;
 }
@@ -38,10 +63,10 @@ pdf_stream_has_crypt(fz_context *ctx, pdf_obj *stm)
 	pdf_obj *obj;
 	int i;
 
-	filters = pdf_dict_geta(ctx, stm, PDF_NAME_Filter, PDF_NAME_F);
+	filters = pdf_dict_geta(ctx, stm, PDF_NAME(Filter), PDF_NAME(F));
 	if (filters)
 	{
-		if (pdf_name_eq(ctx, filters, PDF_NAME_Crypt))
+		if (pdf_name_eq(ctx, filters, PDF_NAME(Crypt)))
 			return 1;
 		if (pdf_is_array(ctx, filters))
 		{
@@ -49,7 +74,7 @@ pdf_stream_has_crypt(fz_context *ctx, pdf_obj *stm)
 			for (i = 0; i < n; i++)
 			{
 				obj = pdf_array_get(ctx, filters, i);
-				if (pdf_name_eq(ctx, obj, PDF_NAME_Crypt))
+				if (pdf_name_eq(ctx, obj, PDF_NAME(Crypt)))
 					return 1;
 			}
 		}
@@ -58,7 +83,7 @@ pdf_stream_has_crypt(fz_context *ctx, pdf_obj *stm)
 }
 
 static fz_jbig2_globals *
-pdf_load_jbig2_globals(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
+pdf_load_jbig2_globals(fz_context *ctx, pdf_obj *dict)
 {
 	fz_jbig2_globals *globals;
 	fz_buffer *buf = NULL;
@@ -93,22 +118,24 @@ pdf_load_jbig2_globals(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 static void
 build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression_params *params)
 {
-	int predictor = pdf_to_int(ctx, pdf_dict_get(ctx, p, PDF_NAME_Predictor));
-	pdf_obj *columns_obj = pdf_dict_get(ctx, p, PDF_NAME_Columns);
+	int predictor = pdf_dict_get_int(ctx, p, PDF_NAME(Predictor));
+	pdf_obj *columns_obj = pdf_dict_get(ctx, p, PDF_NAME(Columns));
 	int columns = pdf_to_int(ctx, columns_obj);
-	int colors = pdf_to_int(ctx, pdf_dict_get(ctx, p, PDF_NAME_Colors));
-	int bpc = pdf_to_int(ctx, pdf_dict_get(ctx, p, PDF_NAME_BitsPerComponent));
+	int colors = pdf_dict_get_int(ctx, p, PDF_NAME(Colors));
+	int bpc = pdf_dict_get_int(ctx, p, PDF_NAME(BitsPerComponent));
+	if (bpc == 0)
+		bpc = 8;
 
 	params->type = FZ_IMAGE_RAW;
 
-	if (pdf_name_eq(ctx, f, PDF_NAME_CCITTFaxDecode) || pdf_name_eq(ctx, f, PDF_NAME_CCF))
+	if (pdf_name_eq(ctx, f, PDF_NAME(CCITTFaxDecode)) || pdf_name_eq(ctx, f, PDF_NAME(CCF)))
 	{
-		pdf_obj *k = pdf_dict_get(ctx, p, PDF_NAME_K);
-		pdf_obj *eol = pdf_dict_get(ctx, p, PDF_NAME_EndOfLine);
-		pdf_obj *eba = pdf_dict_get(ctx, p, PDF_NAME_EncodedByteAlign);
-		pdf_obj *rows = pdf_dict_get(ctx, p, PDF_NAME_Rows);
-		pdf_obj *eob = pdf_dict_get(ctx, p, PDF_NAME_EndOfBlock);
-		pdf_obj *bi1 = pdf_dict_get(ctx, p, PDF_NAME_BlackIs1);
+		pdf_obj *k = pdf_dict_get(ctx, p, PDF_NAME(K));
+		pdf_obj *eol = pdf_dict_get(ctx, p, PDF_NAME(EndOfLine));
+		pdf_obj *eba = pdf_dict_get(ctx, p, PDF_NAME(EncodedByteAlign));
+		pdf_obj *rows = pdf_dict_get(ctx, p, PDF_NAME(Rows));
+		pdf_obj *eob = pdf_dict_get(ctx, p, PDF_NAME(EndOfBlock));
+		pdf_obj *bi1 = pdf_dict_get(ctx, p, PDF_NAME(BlackIs1));
 
 		params->type = FZ_IMAGE_FAX;
 		params->u.fax.k = (k ? pdf_to_int(ctx, k) : 0);
@@ -119,18 +146,18 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
 		params->u.fax.end_of_block = (eob ? pdf_to_bool(ctx, eob) : 1);
 		params->u.fax.black_is_1 = (bi1 ? pdf_to_bool(ctx, bi1) : 0);
 	}
-	else if (pdf_name_eq(ctx, f, PDF_NAME_DCTDecode) || pdf_name_eq(ctx, f, PDF_NAME_DCT))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(DCTDecode)) || pdf_name_eq(ctx, f, PDF_NAME(DCT)))
 	{
-		pdf_obj *ct = pdf_dict_get(ctx, p, PDF_NAME_ColorTransform);
+		pdf_obj *ct = pdf_dict_get(ctx, p, PDF_NAME(ColorTransform));
 
 		params->type = FZ_IMAGE_JPEG;
-		params->u.jpeg.color_transform = (ct ? pdf_to_int(ctx, ct) : -1);
+		params->u.jpeg.color_transform = (ct ? pdf_to_int(ctx, ct) : -2);
 	}
-	else if (pdf_name_eq(ctx, f, PDF_NAME_RunLengthDecode) || pdf_name_eq(ctx, f, PDF_NAME_RL))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(RunLengthDecode)) || pdf_name_eq(ctx, f, PDF_NAME(RL)))
 	{
 		params->type = FZ_IMAGE_RLD;
 	}
-	else if (pdf_name_eq(ctx, f, PDF_NAME_FlateDecode) || pdf_name_eq(ctx, f, PDF_NAME_Fl))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(FlateDecode)) || pdf_name_eq(ctx, f, PDF_NAME(Fl)))
 	{
 		params->type = FZ_IMAGE_FLATE;
 		params->u.flate.predictor = predictor;
@@ -138,9 +165,9 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
 		params->u.flate.colors = colors;
 		params->u.flate.bpc = bpc;
 	}
-	else if (pdf_name_eq(ctx, f, PDF_NAME_LZWDecode) || pdf_name_eq(ctx, f, PDF_NAME_LZW))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(LZWDecode)) || pdf_name_eq(ctx, f, PDF_NAME(LZW)))
 	{
-		pdf_obj *ec = pdf_dict_get(ctx, p, PDF_NAME_EarlyChange);
+		pdf_obj *ec = pdf_dict_get(ctx, p, PDF_NAME(EarlyChange));
 
 		params->type = FZ_IMAGE_LZW;
 		params->u.lzw.predictor = predictor;
@@ -149,6 +176,21 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
 		params->u.lzw.bpc = bpc;
 		params->u.lzw.early_change = (ec ? pdf_to_int(ctx, ec) : 1);
 	}
+	else if (pdf_name_eq(ctx, f, PDF_NAME(JBIG2Decode)))
+	{
+		pdf_obj *g = pdf_dict_get(ctx, p, PDF_NAME(JBIG2Globals));
+
+		params->type = FZ_IMAGE_JBIG2;
+		params->u.jbig2.globals = NULL;
+		params->u.jbig2.embedded = 1; /* jbig2 streams are always embedded without file headers */
+		if (g)
+		{
+			if (!pdf_is_stream(ctx, g))
+				fz_warn(ctx, "jbig2 globals is not a stream, skipping globals");
+			else
+				params->u.jbig2.globals = pdf_load_jbig2_globals(ctx, g);
+		}
+	}
 }
 
 /*
@@ -159,6 +201,7 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 {
 	fz_compression_params local_params;
 
+	local_params.u.jbig2.globals = NULL;
 	if (params == NULL)
 		params = &local_params;
 
@@ -170,39 +213,37 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 	if (params != &local_params && params->type != FZ_IMAGE_RAW)
 		return fz_keep_stream(ctx, chain); /* nothing to do */
 
+	else if (params->type == FZ_IMAGE_JBIG2)
+	{
+		fz_stream *stm;
+		fz_try(ctx)
+			stm = fz_open_image_decomp_stream(ctx, chain, params, NULL);
+		fz_always(ctx)
+			fz_drop_jbig2_globals(ctx, local_params.u.jbig2.globals);
+		fz_catch(ctx)
+			fz_rethrow(ctx);
+		return stm;
+	}
+
 	else if (params->type != FZ_IMAGE_RAW)
 		return fz_open_image_decomp_stream(ctx, chain, params, NULL);
 
-	else if (pdf_name_eq(ctx, f, PDF_NAME_ASCIIHexDecode) || pdf_name_eq(ctx, f, PDF_NAME_AHx))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(ASCIIHexDecode)) || pdf_name_eq(ctx, f, PDF_NAME(AHx)))
 		return fz_open_ahxd(ctx, chain);
 
-	else if (pdf_name_eq(ctx, f, PDF_NAME_ASCII85Decode) || pdf_name_eq(ctx, f, PDF_NAME_A85))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(ASCII85Decode)) || pdf_name_eq(ctx, f, PDF_NAME(A85)))
 		return fz_open_a85d(ctx, chain);
 
-	else if (pdf_name_eq(ctx, f, PDF_NAME_JBIG2Decode))
-	{
-		fz_jbig2_globals *globals = NULL;
-		pdf_obj *obj = pdf_dict_get(ctx, p, PDF_NAME_JBIG2Globals);
-		if (obj)
-		{
-			if (!pdf_is_stream(ctx, obj))
-				fz_warn(ctx, "jbig2 globals is not a stream, skipping globals");
-			else
-				globals = pdf_load_jbig2_globals(ctx, doc, obj);
-		}
-		return fz_open_jbig2d(ctx, chain, globals); /* takes ownership of jbig2_globals */
-	}
-
-	else if (pdf_name_eq(ctx, f, PDF_NAME_JPXDecode))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(JPXDecode)))
 		return fz_keep_stream(ctx, chain); /* JPX decoding is special cased in the image loading code */
 
-	else if (pdf_name_eq(ctx, f, PDF_NAME_Crypt))
+	else if (pdf_name_eq(ctx, f, PDF_NAME(Crypt)))
 	{
 		if (!doc->crypt)
 			fz_warn(ctx, "crypt filter in unencrypted document");
 		else
 		{
-			pdf_obj *name = pdf_dict_get(ctx, p, PDF_NAME_Name);
+			pdf_obj *name = pdf_dict_get(ctx, p, PDF_NAME(Name));
 			if (pdf_is_name(ctx, name))
 				return pdf_open_crypt_with_filter(ctx, chain, doc->crypt, name, num, gen);
 		}
@@ -248,10 +289,7 @@ build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pd
 		}
 	}
 	fz_catch(ctx)
-	{
-		fz_drop_stream(ctx, chain);
 		fz_rethrow(ctx);
-	}
 	return chain;
 }
 
@@ -280,22 +318,27 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 	if (num > 0 && num < pdf_xref_len(ctx, doc))
 	{
 		x = pdf_get_xref_entry(ctx, doc, num);
-		*orig_num = x->num;
-		*orig_gen = x->gen;
-		if (x->stm_buf)
-			return fz_open_buffer(ctx, x->stm_buf);
 	}
-	else
+	if (x == NULL)
 	{
 		/* We only end up here when called from pdf_open_stream_with_offset to parse new format XRef sections. */
 		/* New style XRef sections must have generation number 0. */
 		*orig_num = num;
 		*orig_gen = 0;
 	}
+	else
+	{
+		*orig_num = x->num;
+		*orig_gen = x->gen;
+		if (x->stm_buf)
+			return fz_open_buffer(ctx, x->stm_buf);
+	}
 
 	hascrypt = pdf_stream_has_crypt(ctx, stmobj);
-	len = pdf_to_int(ctx, pdf_dict_get(ctx, stmobj, PDF_NAME_Length));
-	null_stm = fz_open_null(ctx, file_stm, len, offset);
+	len = pdf_dict_get_int(ctx, stmobj, PDF_NAME(Length));
+	if (len < 0)
+		len = 0;
+	null_stm = fz_open_endstream_filter(ctx, file_stm, len, offset);
 	if (doc->crypt && !hascrypt)
 	{
 		fz_try(ctx)
@@ -316,8 +359,8 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 static fz_stream *
 pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj *stmobj, int num, int64_t offset, fz_compression_params *imparams)
 {
-	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME_Filter, PDF_NAME_F);
-	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME_DecodeParms, PDF_NAME_DP);
+	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME(Filter), PDF_NAME(F));
+	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
 	int orig_num, orig_gen;
 	fz_stream *rstm, *fstm;
 
@@ -329,7 +372,11 @@ pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj
 		else if (pdf_array_len(ctx, filters) > 0)
 			fstm = build_filter_chain(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams);
 		else
+		{
+			if (imparams)
+				imparams->type = FZ_IMAGE_RAW;
 			fstm = fz_keep_stream(ctx, rstm);
+		}
 	}
 	fz_always(ctx)
 		fz_drop_stream(ctx, rstm);
@@ -339,15 +386,11 @@ pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj
 	return fstm;
 }
 
-/*
- * Construct a filter to decode a stream, without
- * constraining to stream length, and without decryption.
- */
 fz_stream *
 pdf_open_inline_stream(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj, int length, fz_stream *file_stm, fz_compression_params *imparams)
 {
-	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME_Filter, PDF_NAME_F);
-	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME_DecodeParms, PDF_NAME_DP);
+	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME(Filter), PDF_NAME(F));
+	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
 
 	if (pdf_is_name(ctx, filters))
 		return build_filter(ctx, file_stm, doc, filters, params, 0, 0, imparams);
@@ -356,7 +399,7 @@ pdf_open_inline_stream(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj, int
 
 	if (imparams)
 		imparams->type = FZ_IMAGE_RAW;
-	return fz_open_null(ctx, file_stm, length, fz_tell(ctx, file_stm));
+	return fz_open_null_filter(ctx, file_stm, length, fz_tell(ctx, file_stm));
 }
 
 void
@@ -379,8 +422,7 @@ pdf_load_compressed_inline_image(fz_context *ctx, pdf_document *doc, pdf_obj *di
 		istm = pdf_open_inline_stream(ctx, doc, dict, length, file_stm, &bc->params);
 		leech = fz_open_leecher(ctx, istm, bc->buffer);
 		decomp = fz_open_image_decomp_stream(ctx, leech, &bc->params, &dummy_l2factor);
-		pixmap = fz_decomp_image_from_stream(ctx, decomp, image, NULL, indexed, 0);
-		fz_set_compressed_image_tile(ctx, image, pixmap);
+		pixmap = fz_decomp_image_from_stream(ctx, decomp, image, NULL, indexed, 0, NULL);
 		fz_set_compressed_image_buffer(ctx, image, bc);
 	}
 	fz_always(ctx)
@@ -397,18 +439,12 @@ pdf_load_compressed_inline_image(fz_context *ctx, pdf_document *doc, pdf_obj *di
 	}
 }
 
-/*
- * Open a stream for reading the raw (compressed but decrypted) data.
- */
 fz_stream *
 pdf_open_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
 	pdf_xref_entry *x;
 	int orig_num, orig_gen;
 
-	if (num <= 0 || num >= pdf_xref_len(ctx, doc))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "object id out of range (%d 0 R)", num);
-
 	x = pdf_cache_object(ctx, doc, num);
 	if (x->stm_ofs == 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
@@ -421,9 +457,6 @@ pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 {
 	pdf_xref_entry *x;
 
-	if (num <= 0 || num >= pdf_xref_len(ctx, doc))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "object id out of range (%d 0 R)", num);
-
 	x = pdf_cache_object(ctx, doc, num);
 	if (x->stm_ofs == 0 && x->stm_buf == NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
@@ -431,11 +464,6 @@ pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	return pdf_open_filter(ctx, doc, doc->file, x->obj, num, x->stm_ofs, params);
 }
 
-/*
- * Open a stream for reading uncompressed data.
- * Put the opened file in doc->stream.
- * Using doc->file while a stream is open is a Bad idea.
- */
 fz_stream *
 pdf_open_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
@@ -450,9 +478,6 @@ pdf_open_stream_with_offset(fz_context *ctx, pdf_document *doc, int num, pdf_obj
 	return pdf_open_filter(ctx, doc, doc->file, dict, num, stm_ofs, NULL);
 }
 
-/*
- * Load raw (compressed but decrypted) contents of a stream into buf.
- */
 fz_buffer *
 pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
@@ -464,16 +489,19 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 
 	if (num > 0 && num < pdf_xref_len(ctx, doc))
 	{
-		x = pdf_get_xref_entry(ctx, doc, num);
+		x = pdf_get_xref_entry_no_null(ctx, doc, num);
 		if (x->stm_buf)
 			return fz_keep_buffer(ctx, x->stm_buf);
 	}
 
 	dict = pdf_load_object(ctx, doc, num);
 
-	len = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_Length));
-
-	pdf_drop_obj(ctx, dict);
+	fz_try(ctx)
+		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+	fz_always(ctx)
+		pdf_drop_obj(ctx, dict);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 
 	stm = pdf_open_raw_stream_number(ctx, doc, num);
 
@@ -490,6 +518,8 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 static int
 pdf_guess_filter_length(int len, const char *filter)
 {
+	if (len < 0)
+		len = 0;
 	if (!strcmp(filter, "ASCIIHexDecode"))
 		return len / 2;
 	if (!strcmp(filter, "ASCII85Decode"))
@@ -522,12 +552,12 @@ can_reuse_buffer(fz_context *ctx, pdf_xref_entry *entry, fz_compression_params *
 	if (params)
 		params->type = FZ_IMAGE_RAW;
 
-	f = pdf_dict_geta(ctx, entry->obj, PDF_NAME_Filter, PDF_NAME_F);
+	f = pdf_dict_geta(ctx, entry->obj, PDF_NAME(Filter), PDF_NAME(F));
 	/* If there are no filters, it's uncompressed, and we can use it */
 	if (!f)
 		return 1;
 
-	p = pdf_dict_geta(ctx, entry->obj, PDF_NAME_DecodeParms, PDF_NAME_DP);
+	p = pdf_dict_geta(ctx, entry->obj, PDF_NAME(DecodeParms), PDF_NAME(DP));
 	if (pdf_is_array(ctx, f))
 	{
 		int len = pdf_array_len(ctx, f);
@@ -556,7 +586,7 @@ can_reuse_buffer(fz_context *ctx, pdf_xref_entry *entry, fz_compression_params *
 }
 
 static fz_buffer *
-pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int *truncated)
+pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int *truncated, size_t worst_case)
 {
 	fz_stream *stm = NULL;
 	pdf_obj *dict, *obj;
@@ -575,24 +605,29 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	}
 
 	dict = pdf_load_object(ctx, doc, num);
-
-	len = pdf_to_int(ctx, pdf_dict_get(ctx, dict, PDF_NAME_Length));
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_Filter);
-	len = pdf_guess_filter_length(len, pdf_to_name(ctx, obj));
-	n = pdf_array_len(ctx, obj);
-	for (i = 0; i < n; i++)
-		len = pdf_guess_filter_length(len, pdf_to_name(ctx, pdf_array_get(ctx, obj, i)));
-
-	pdf_drop_obj(ctx, dict);
+	fz_try(ctx)
+	{
+		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+		obj = pdf_dict_get(ctx, dict, PDF_NAME(Filter));
+		len = pdf_guess_filter_length(len, pdf_to_name(ctx, obj));
+		n = pdf_array_len(ctx, obj);
+		for (i = 0; i < n; i++)
+			len = pdf_guess_filter_length(len, pdf_to_name(ctx, pdf_array_get(ctx, obj, i)));
+	}
+	fz_always(ctx)
+	{
+		pdf_drop_obj(ctx, dict);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
 
 	stm = pdf_open_image_stream(ctx, doc, num, params);
 
 	fz_try(ctx)
 	{
-		if (truncated)
-			buf = fz_read_best(ctx, stm, len, truncated);
-		else
-			buf = fz_read_all(ctx, stm, len);
+		buf = fz_read_best(ctx, stm, len, truncated, worst_case);
 	}
 	fz_always(ctx)
 	{
@@ -606,29 +641,20 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	return buf;
 }
 
-/*
- * Load uncompressed contents of a stream into buf.
- */
 fz_buffer *
 pdf_load_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
-	return pdf_load_image_stream(ctx, doc, num, NULL, NULL);
-}
-
-fz_buffer *
-pdf_load_stream_truncated(fz_context *ctx, pdf_document *doc, int num, int *truncated)
-{
-	return pdf_load_image_stream(ctx, doc, num, NULL, truncated);
+	return pdf_load_image_stream(ctx, doc, num, NULL, NULL, 0);
 }
 
 fz_compressed_buffer *
-pdf_load_compressed_stream(fz_context *ctx, pdf_document *doc, int num)
+pdf_load_compressed_stream(fz_context *ctx, pdf_document *doc, int num, size_t worst_case)
 {
 	fz_compressed_buffer *bc = fz_malloc_struct(ctx, fz_compressed_buffer);
 
 	fz_try(ctx)
 	{
-		bc->buffer = pdf_load_image_stream(ctx, doc, num, &bc->params, NULL);
+		bc->buffer = pdf_load_image_stream(ctx, doc, num, &bc->params, NULL, worst_case);
 	}
 	fz_catch(ctx)
 	{
@@ -678,7 +704,8 @@ pdf_open_contents_stream(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 	if (pdf_is_stream(ctx, obj))
 		return pdf_open_image_stream(ctx, doc, num, NULL);
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "pdf object stream missing (%d 0 R)", num);
+	fz_warn(ctx, "content stream is not a stream (%d 0 R)", num);
+	return fz_open_memory(ctx, (unsigned char *)"", 0);
 }
 
 fz_buffer *pdf_load_raw_stream(fz_context *ctx, pdf_obj *ref)
