@@ -1,7 +1,7 @@
 //---------------------------------------------------------------------------------
 //
 //  Little Color Management System
-//  Copyright (c) 1998-2017 Marti Maria Saguer
+//  Copyright (c) 1998-2022 Marti Maria Saguer
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the "Software"),
@@ -129,13 +129,22 @@ typedef struct _Table {
 
 // File stream being parsed
 typedef struct _FileContext {
-        char           FileName[cmsMAX_PATH];    // File name if being readed from file
+        char           FileName[cmsMAX_PATH];    // File name if being read from file
         FILE*          Stream;                   // File stream or NULL if holded in memory
     } FILECTX;
 
-// This struct hold all information about an open IT8 handler.
+//Very simple string
 typedef struct {
 
+        struct struct_it8* it8;
+        cmsInt32Number max;
+        cmsInt32Number len;
+        char* begin;
+    } string;
+
+
+// This struct hold all information about an open IT8 handler.
+typedef struct struct_it8 {
 
         cmsUInt32Number  TablesCount;                     // How many tables in this stream
         cmsUInt32Number  nTable;                          // The actual table
@@ -153,8 +162,8 @@ typedef struct {
         cmsInt32Number     inum;              // integer value
         cmsFloat64Number   dnum;              // real value
 
-        char           id[MAXID];             // identifier
-        char           str[MAXSTR];           // string
+        string*        id;            // identifier
+        string*        str;           // string
 
         // Allowed keywords & datasets. They have visibility on whole stream
         KEYVALUE*      ValidKeywords;
@@ -237,7 +246,7 @@ static PROPERTY PredefinedProperties[] = {
 
         {"MATERIAL",         WRITE_STRINGIFY},    // Identifies the material on which the target was produced using a code
                                                   // uniquely identifying th e material. This is intend ed to be used for IT8.7
-                                                  // physical targets only (i.e . IT8.7/1 a nd IT8.7/2).
+                                                  // physical targets only (i.e . IT8.7/1 and IT8.7/2).
 
         {"INSTRUMENTATION",  WRITE_STRINGIFY},    // Used to report the specific instrumentation used (manufacturer and
                                                   // model number) to generate the data reported. This data will often
@@ -254,7 +263,7 @@ static PROPERTY PredefinedProperties[] = {
                                                    // needed.
 
         {"SAMPLE_BACKING",   WRITE_STRINGIFY},     // Identifies the backing material used behind the sample during
-                                                   // measurement. Allowed values are black, white, or {"na".
+                                                   // measurement. Allowed values are "black", "white", or {"na".
 
         {"CHISQ_DOF",        WRITE_STRINGIFY},     // Degrees of freedom associated with the Chi squared statistic
                                                    // below properties are new in recent specs:
@@ -269,7 +278,7 @@ static PROPERTY PredefinedProperties[] = {
                                                    // denote the use of filters such as none, D65, Red, Green or Blue.
 
        {"POLARIZATION",      WRITE_STRINGIFY},     // Identifies the use of a physical polarization filter during measurement. Allowed
-                                                   // values are {"yes, white, none or na.
+                                                   // values are {"yes", "white", "none" or "na".
 
        {"WEIGHTING_FUNCTION", WRITE_PAIR},         // Indicates such functions as: the CIE standard observer functions used in the
                                                    // calculation of various data parameters (2 degree and 10 degree), CIE standard
@@ -352,6 +361,65 @@ static const char* PredefinedSampleID[] = {
 //Forward declaration of some internal functions
 static void* AllocChunk(cmsContext ContextID, cmsIT8* it8, cmsUInt32Number size);
 
+static
+string* StringAlloc(cmsContext ContextID, cmsIT8* it8, int max)
+{
+    string* s = (string*) AllocChunk(ContextID, it8, sizeof(string));
+    if (s == NULL) return NULL;
+
+    s->it8 = it8;
+    s->max = max;
+    s->len = 0;
+    s->begin = (char*) AllocChunk(ContextID, it8, s->max);
+
+    return s;
+}
+
+static
+void StringClear(string* s)
+{
+    s->len = 0;
+}
+
+static
+void StringAppend(cmsContext ContextID, string* s, char c)
+{
+    if (s->len + 1 >= s->max)
+    {
+        char* new_ptr;
+
+        s->max *= 10;
+        new_ptr = (char*) AllocChunk(ContextID, s->it8, s->max);
+        if (new_ptr != NULL && s->begin != NULL)
+            memcpy(new_ptr, s->begin, s->len);
+
+        s->begin = new_ptr;
+    }
+
+    if (s->begin != NULL)
+    {
+        s->begin[s->len++] = c;
+        s->begin[s->len] = 0;
+    }
+}
+
+static
+char* StringPtr(string* s)
+{
+    return s->begin;
+}
+
+static
+void StringCat(cmsContext ContextID, string* s, const char* c)
+{
+    while (*c)
+    {
+        StringAppend(ContextID, s, *c);
+        c++;
+    }
+}
+
+
 // Checks whatever c is a separator
 static
 cmsBool isseparator(int c)
@@ -478,10 +546,8 @@ cmsBool Check(cmsContext ContextID, cmsIT8* it8, SYMBOL sy, const char* Err)
 
 // Read Next character from stream
 static
-void NextCh(cmsContext ContextID, cmsIT8* it8)
+void NextCh(cmsIT8* it8)
 {
-    cmsUNUSED_PARAMETER(ContextID);
-
     if (it8 -> FileStack[it8 ->IncludeSP]->Stream) {
 
         it8 ->ch = fgetc(it8 ->FileStack[it8 ->IncludeSP]->Stream);
@@ -506,12 +572,11 @@ void NextCh(cmsContext ContextID, cmsIT8* it8)
 
 // Try to see if current identifier is a keyword, if so return the referred symbol
 static
-SYMBOL BinSrchKey(cmsContext ContextID, const char *id)
+SYMBOL BinSrchKey(const char *id)
 {
     int l = 1;
     int r = NUMKEYS;
     int x, res;
-    cmsUNUSED_PARAMETER(ContextID);
 
     while (r >= l)
     {
@@ -536,14 +601,14 @@ cmsFloat64Number xpow10(int n)
 
 //  Reads a Real number, tries to follow from integer number
 static
-void ReadReal(cmsContext ContextID, cmsIT8* it8, cmsInt32Number inum)
+void ReadReal(cmsIT8* it8, cmsInt32Number inum)
 {
     it8->dnum = (cmsFloat64Number)inum;
 
     while (isdigit(it8->ch)) {
 
         it8->dnum = (cmsFloat64Number)it8->dnum * 10.0 + (cmsFloat64Number)(it8->ch - '0');
-        NextCh(ContextID, it8);
+        NextCh(it8);
     }
 
     if (it8->ch == '.') {        // Decimal point
@@ -551,13 +616,13 @@ void ReadReal(cmsContext ContextID, cmsIT8* it8, cmsInt32Number inum)
         cmsFloat64Number frac = 0.0;      // fraction
         int prec = 0;                     // precision
 
-        NextCh(ContextID, it8);               // Eats dec. point
+        NextCh(it8);               // Eats dec. point
 
         while (isdigit(it8->ch)) {
 
             frac = frac * 10.0 + (cmsFloat64Number)(it8->ch - '0');
             prec++;
-            NextCh(ContextID, it8);
+            NextCh(it8);
         }
 
         it8->dnum = it8->dnum + (frac / xpow10(prec));
@@ -569,17 +634,17 @@ void ReadReal(cmsContext ContextID, cmsIT8* it8, cmsInt32Number inum)
         cmsInt32Number e;
         cmsInt32Number sgn;
 
-        NextCh(ContextID, it8); sgn = 1;
+        NextCh(it8); sgn = 1;
 
         if (it8->ch == '-') {
 
-            sgn = -1; NextCh(ContextID, it8);
+            sgn = -1; NextCh(it8);
         }
         else
             if (it8->ch == '+') {
 
                 sgn = +1;
-                NextCh(ContextID, it8);
+                NextCh(it8);
             }
 
         e = 0;
@@ -590,7 +655,7 @@ void ReadReal(cmsContext ContextID, cmsIT8* it8, cmsInt32Number inum)
             if ((cmsFloat64Number)e * 10.0 + (cmsFloat64Number)digit < (cmsFloat64Number)+2147483647.0)
                 e = e * 10 + digit;
 
-            NextCh(ContextID, it8);
+            NextCh(it8);
         }
 
         e = sgn*e;
@@ -602,11 +667,10 @@ void ReadReal(cmsContext ContextID, cmsIT8* it8, cmsInt32Number inum)
 // This can not call directly atof because it uses locale dependent
 // parsing, while CCMX files always use . as decimal separator
 static
-cmsFloat64Number ParseFloatNumber(cmsContext ContextID, const char *Buffer)
+cmsFloat64Number ParseFloatNumber(const char *Buffer)
 {
     cmsFloat64Number dnum = 0.0;
     int sign = 1;
-    cmsUNUSED_PARAMETER(ContextID);
 
     // keep safe
     if (Buffer == NULL) return 0.0;
@@ -681,37 +745,64 @@ cmsFloat64Number ParseFloatNumber(cmsContext ContextID, const char *Buffer)
 }
 
 
+// Reads a string, special case to avoid infinite resursion on .include
+static
+void InStringSymbol(cmsContext ContextID, cmsIT8* it8)
+{
+    while (isseparator(it8->ch))
+        NextCh(it8);
+
+    if (it8->ch == '\'' || it8->ch == '\"')
+    {
+        int sng;
+
+        sng = it8->ch;
+        StringClear(it8->str);
+
+        NextCh(it8);
+
+        while (it8->ch != sng) {
+
+            if (it8->ch == '\n' || it8->ch == '\r' || it8->ch == 0) break;
+            else {
+                StringAppend(ContextID, it8->str, (char)it8->ch);
+                NextCh(it8);
+            }
+        }
+
+        it8->sy = SSTRING;
+        NextCh(it8);
+    }
+    else
+        SynError(ContextID, it8, "String expected");
+
+}
+
 // Reads next symbol
 static
 void InSymbol(cmsContext ContextID, cmsIT8* it8)
 {
-    register char *idptr;
-    register int k;
     SYMBOL key;
-    int sng;
 
     do {
 
         while (isseparator(it8->ch))
-            NextCh(ContextID, it8);
+            NextCh(it8);
 
         if (isfirstidchar(it8->ch)) {          // Identifier
 
-            k = 0;
-            idptr = it8->id;
+            StringClear(it8->id);
 
             do {
 
-                if (++k < MAXID) *idptr++ = (char) it8->ch;
+                StringAppend(ContextID, it8->id, (char) it8->ch);
 
-                NextCh(ContextID, it8);
+                NextCh(it8);
 
             } while (isidchar(it8->ch));
 
-            *idptr = '\0';
 
-
-            key = BinSrchKey(ContextID, it8->id);
+            key = BinSrchKey(StringPtr(it8->id));
             if (key == SUNDEFINED) it8->sy = SIDENT;
             else it8->sy = key;
 
@@ -723,7 +814,7 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
 
                 if (it8->ch == '-') {
                     sign = -1;
-                    NextCh(ContextID, it8);
+                    NextCh(it8);
                 }
 
                 it8->inum = 0;
@@ -731,12 +822,12 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
 
                 if (it8->ch == '0') {          // 0xnnnn (Hexa) or 0bnnnn (Binary)
 
-                    NextCh(ContextID, it8);
+                    NextCh(it8);
                     if (toupper(it8->ch) == 'X') {
 
                         int j;
 
-                        NextCh(ContextID, it8);
+                        NextCh(it8);
                         while (isxdigit(it8->ch))
                         {
                             it8->ch = toupper(it8->ch);
@@ -746,11 +837,12 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
                             if ((cmsFloat64Number) it8->inum * 16.0 + (cmsFloat64Number) j > (cmsFloat64Number)+2147483647.0)
                             {
                                 SynError(ContextID, it8, "Invalid hexadecimal number");
+                                it8->sy = SEOF;
                                 return;
                             }
 
                             it8->inum = it8->inum * 16 + j;
-                            NextCh(ContextID, it8);
+                            NextCh(it8);
                         }
                         return;
                     }
@@ -759,7 +851,7 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
 
                         int j;
 
-                        NextCh(ContextID, it8);
+                        NextCh(it8);
                         while (it8->ch == '0' || it8->ch == '1')
                         {
                             j = it8->ch - '0';
@@ -767,11 +859,12 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
                             if ((cmsFloat64Number) it8->inum * 2.0 + j > (cmsFloat64Number)+2147483647.0)
                             {
                                 SynError(ContextID, it8, "Invalid binary number");
+                                it8->sy = SEOF;
                                 return;
                             }
 
                             it8->inum = it8->inum * 2 + j;
-                            NextCh(ContextID, it8);
+                            NextCh(it8);
                         }
                         return;
                     }
@@ -783,19 +876,19 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
                     cmsInt32Number digit = (it8->ch - '0');
 
                     if ((cmsFloat64Number) it8->inum * 10.0 + (cmsFloat64Number) digit > (cmsFloat64Number) +2147483647.0) {
-                        ReadReal(ContextID, it8, it8->inum);
+                        ReadReal(it8, it8->inum);
                         it8->sy = SDNUM;
                         it8->dnum *= sign;
                         return;
                     }
 
                     it8->inum = it8->inum * 10 + digit;
-                    NextCh(ContextID, it8);
+                    NextCh(it8);
                 }
 
                 if (it8->ch == '.') {
 
-                    ReadReal(ContextID, it8, it8->inum);
+                    ReadReal(it8, it8->inum);
                     it8->sy = SDNUM;
                     it8->dnum *= sign;
                     return;
@@ -807,26 +900,27 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
 
                 if (isidchar(it8 ->ch)) {
 
+                    char buffer[127];
+
                     if (it8 ->sy == SINUM) {
 
-                        snprintf(it8->id, 127, "%d", it8->inum);
+                        snprintf(buffer, sizeof(buffer), "%d", it8->inum);
                     }
                     else {
 
-                        snprintf(it8->id, 127, it8 ->DoubleFormatter, it8->dnum);
+                        snprintf(buffer, sizeof(buffer), it8 ->DoubleFormatter, it8->dnum);
                     }
 
-                    k = (int) strlen(it8 ->id);
-                    idptr = it8 ->id + k;
+                    StringCat(ContextID, it8->id, buffer);
+
                     do {
 
-                        if (++k < MAXID) *idptr++ = (char) it8->ch;
+                        StringAppend(ContextID, it8->id, (char) it8->ch);
 
-                        NextCh(ContextID, it8);
+                        NextCh(it8);
 
                     } while (isidchar(it8->ch));
 
-                    *idptr = '\0';
                     it8->sy = SIDENT;
                 }
                 return;
@@ -835,12 +929,8 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
             else
                 switch ((int) it8->ch) {
 
-        // EOF marker -- ignore it
-        case '\x1a':
-            NextCh(ContextID, it8);
-            break;
-
         // Eof stream markers
+        case '\x1a':
         case 0:
         case -1:
             it8->sy = SEOF;
@@ -849,24 +939,24 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
 
         // Next line
         case '\r':
-            NextCh(ContextID, it8);
+            NextCh(it8);
             if (it8 ->ch == '\n')
-                NextCh(ContextID, it8);
+                NextCh(it8);
             it8->sy = SEOLN;
             it8->lineno++;
             break;
 
         case '\n':
-            NextCh(ContextID, it8);
+            NextCh(it8);
             it8->sy = SEOLN;
             it8->lineno++;
             break;
 
         // Comment
         case '#':
-            NextCh(ContextID, it8);
+            NextCh(it8);
             while (it8->ch && it8->ch != '\n' && it8->ch != '\r')
-                NextCh(ContextID, it8);
+                NextCh(it8);
 
             it8->sy = SCOMMENT;
             break;
@@ -874,29 +964,13 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
         // String.
         case '\'':
         case '\"':
-            idptr = it8->str;
-            sng = it8->ch;
-            k = 0;
-            NextCh(ContextID, it8);
-
-            while (k < (MAXSTR-1) && it8->ch != sng) {
-
-                if (it8->ch == '\n'|| it8->ch == '\r') k = MAXSTR+1;
-                else {
-                    *idptr++ = (char) it8->ch;
-                    NextCh(ContextID, it8);
-                    k++;
-                }
-            }
-
-            it8->sy = SSTRING;
-            *idptr = '\0';
-            NextCh(ContextID, it8);
+            InStringSymbol(ContextID, it8);
             break;
 
 
         default:
             SynError(ContextID, it8, "Unrecognized character: 0x%x", it8 ->ch);
+            it8->sy = SEOF;
             return;
             }
 
@@ -911,24 +985,33 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
                 if(it8 -> IncludeSP >= (MAXINCLUDE-1)) {
 
                     SynError(ContextID, it8, "Too many recursion levels");
+                    it8->sy = SEOF;
                     return;
                 }
 
-                InSymbol(ContextID, it8);
-                if (!Check(ContextID, it8, SSTRING, "Filename expected")) return;
+                InStringSymbol(ContextID, it8);
+                if (!Check(ContextID, it8, SSTRING, "Filename expected"))
+                {
+                    it8->sy = SEOF;
+                    return;
+                }
 
                 FileNest = it8 -> FileStack[it8 -> IncludeSP + 1];
                 if(FileNest == NULL) {
 
                     FileNest = it8 ->FileStack[it8 -> IncludeSP + 1] = (FILECTX*)AllocChunk(ContextID, it8, sizeof(FILECTX));
-                    //if(FileNest == NULL)
-                    //  TODO: how to manage out-of-memory conditions?
+                    if (FileNest == NULL) {
+                        SynError(ContextID, it8, "Out of memory");
+                        it8->sy = SEOF;
+                        return;
+                    }
                 }
 
-                if (BuildAbsolutePath(it8->str,
+                if (BuildAbsolutePath(StringPtr(it8->str),
                                       it8->FileStack[it8->IncludeSP]->FileName,
                                       FileNest->FileName, cmsMAX_PATH-1) == FALSE) {
                     SynError(ContextID, it8, "File path too long");
+                    it8->sy = SEOF;
                     return;
                 }
 
@@ -936,6 +1019,7 @@ void InSymbol(cmsContext ContextID, cmsIT8* it8)
                 if (FileNest->Stream == NULL) {
 
                         SynError(ContextID, it8, "File %s not found", FileNest->FileName);
+                        it8->sy = SEOF;
                         return;
                 }
                 it8->IncludeSP++;
@@ -986,12 +1070,12 @@ cmsBool GetVal(cmsContext ContextID, cmsIT8* it8, char* Buffer, cmsUInt32Number
     case SEOLN:   // Empty value
                   Buffer[0]=0;
                   break;
-    case SIDENT:  strncpy(Buffer, it8->id, max);
+    case SIDENT:  strncpy(Buffer, StringPtr(it8->id), max);
                   Buffer[max-1]=0;
                   break;
     case SINUM:   snprintf(Buffer, max, "%d", it8 -> inum); break;
     case SDNUM:   snprintf(Buffer, max, it8->DoubleFormatter, it8 -> dnum); break;
-    case SSTRING: strncpy(Buffer, it8->str, max);
+    case SSTRING: strncpy(Buffer, StringPtr(it8->str), max);
                   Buffer[max-1] = 0;
                   break;
 
@@ -1099,6 +1183,9 @@ void* AllocChunk(cmsContext ContextID, cmsIT8* it8, cmsUInt32Number size)
         it8 ->Allocator.Block = (cmsUInt8Number*)  AllocBigBlock(ContextID, it8, it8 ->Allocator.BlockSize);
     }
 
+    if (it8->Allocator.Block == NULL)
+        return NULL;
+
     ptr = it8 ->Allocator.Block + it8 ->Allocator.Used;
     it8 ->Allocator.Used += size;
 
@@ -1116,7 +1203,7 @@ char *AllocString(cmsContext ContextID, cmsIT8* it8, const char* str)
 
 
     ptr = (char *) AllocChunk(ContextID, it8, Size);
-    if (ptr) strncpy (ptr, str, Size-1);
+    if (ptr) memcpy(ptr, str, Size-1);
 
     return ptr;
 }
@@ -1316,6 +1403,9 @@ cmsHANDLE  CMSEXPORT cmsIT8Alloc(cmsContext ContextID)
     it8->IncludeSP   = 0;
     it8 -> lineno = 1;
 
+    it8->id = StringAlloc(ContextID, it8, MAXSTR);
+    it8->str = StringAlloc(ContextID, it8, MAXSTR);
+
     strcpy(it8->DoubleFormatter, DEFAULT_DBL_FORMAT);
     cmsIT8SetSheetType(ContextID, (cmsHANDLE) it8, "CGATS.17");
 
@@ -1421,7 +1511,7 @@ cmsFloat64Number CMSEXPORT cmsIT8GetPropertyDbl(cmsContext ContextID, cmsHANDLE
 
     if (v == NULL) return 0.0;
 
-    return ParseFloatNumber(ContextID, v);
+    return ParseFloatNumber(v);
 }
 
 const char* CMSEXPORT cmsIT8GetPropertyMulti(cmsContext ContextID, cmsHANDLE hIT8, const char* Key, const char *SubKey)
@@ -1437,28 +1527,45 @@ const char* CMSEXPORT cmsIT8GetPropertyMulti(cmsContext ContextID, cmsHANDLE hIT
 
 // ----------------------------------------------------------------- Datasets
 
+// A safe atoi that returns 0 when NULL input is given
+static
+cmsInt32Number satoi(const char* b)
+{
+    int n;
+
+    if (b == NULL) return 0;
+
+    n = atoi(b);
+    if (n > 0x7fffffffL) return 0x7fffffffL;
+    if (n < -0x7ffffffeL) return -0x7ffffffeL;
+
+    return (cmsInt32Number)n;
+}
+
 
 static
-void AllocateDataFormat(cmsContext ContextID, cmsIT8* it8)
+cmsBool AllocateDataFormat(cmsContext ContextID, cmsIT8* it8)
 {
     TABLE* t = GetTable(ContextID, it8);
 
-    if (t -> DataFormat) return;    // Already allocated
+    if (t -> DataFormat) return TRUE;    // Already allocated
 
-    t -> nSamples  = (int) cmsIT8GetPropertyDbl(ContextID, it8, "NUMBER_OF_FIELDS");
+    t -> nSamples  = satoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_FIELDS"));
 
     if (t -> nSamples <= 0) {
 
         SynError(ContextID, it8, "AllocateDataFormat: Unknown NUMBER_OF_FIELDS");
-        t -> nSamples = 10;
-        }
+        return FALSE;
+    }
 
     t -> DataFormat = (char**) AllocChunk (ContextID, it8, ((cmsUInt32Number) t->nSamples + 1) * sizeof(char *));
     if (t->DataFormat == NULL) {
 
         SynError(ContextID, it8, "AllocateDataFormat: Unable to allocate dataFormat array");
+        return FALSE;
     }
 
+    return TRUE;
 }
 
 static
@@ -1477,8 +1584,11 @@ cmsBool SetDataFormat(cmsContext ContextID, cmsIT8* it8, int n, const char *labe
 {
     TABLE* t = GetTable(ContextID, it8);
 
-    if (!t->DataFormat)
-        AllocateDataFormat(ContextID, it8);
+    if (!t->DataFormat) {
+
+        if (!AllocateDataFormat(ContextID, it8))
+            return FALSE;
+    }
 
     if (n > t -> nSamples) {
         SynError(ContextID, it8, "More than NUMBER_OF_FIELDS fields.");
@@ -1487,6 +1597,7 @@ cmsBool SetDataFormat(cmsContext ContextID, cmsIT8* it8, int n, const char *labe
 
     if (t->DataFormat) {
         t->DataFormat[n] = AllocString(ContextID, it8, label);
+        if (t->DataFormat[n] == NULL) return FALSE;
     }
 
     return TRUE;
@@ -1499,22 +1610,51 @@ cmsBool CMSEXPORT cmsIT8SetDataFormat(cmsContext ContextID, cmsHANDLE  h, int n,
     return SetDataFormat(ContextID, it8, n, Sample);
 }
 
+// Convert to binary
 static
-void AllocateDataSet(cmsContext ContextID, cmsIT8* it8)
+const char* satob(const char* v)
+{
+    cmsUInt32Number x;
+    static char buf[33];
+    char *s = buf + 33;
+
+    if (v == NULL) return "0";
+
+    x = atoi(v);
+    *--s = 0;
+    if (!x) *--s = '0';
+    for (; x; x /= 2) *--s = '0' + x%2;
+
+    return s;
+}
+
+
+static
+cmsBool AllocateDataSet(cmsContext ContextID, cmsIT8* it8)
 {
     TABLE* t = GetTable(ContextID, it8);
 
-    if (t -> Data) return;    // Already allocated
+    if (t -> Data) return TRUE;    // Already allocated
 
-    t-> nSamples   = atoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_FIELDS"));
-    t-> nPatches   = atoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_SETS"));
+    t-> nSamples   = satoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_FIELDS"));
+    t-> nPatches   = satoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_SETS"));
 
-    t-> Data = (char**)AllocChunk (ContextID, it8, ((cmsUInt32Number) t->nSamples + 1) * ((cmsUInt32Number) t->nPatches + 1) *sizeof (char*));
-    if (t->Data == NULL) {
+    if (t -> nSamples < 0 || t->nSamples > 0x7ffe || t->nPatches < 0 || t->nPatches > 0x7ffe)
+    {
+        SynError(ContextID, it8, "AllocateDataSet: too much data");
+        return FALSE;
+    }
+    else {
+        // Some dumb analizers warns of possible overflow here, just take a look couple of lines above.
+        t->Data = (char**)AllocChunk(ContextID, it8, ((cmsUInt32Number)t->nSamples + 1) * ((cmsUInt32Number)t->nPatches + 1) * sizeof(char*));
+        if (t->Data == NULL) {
 
-        SynError(ContextID, it8, "AllocateDataSet: Unable to allocate data array");
+            SynError(ContextID, it8, "AllocateDataSet: Unable to allocate data array");
+            return FALSE;
+        }
     }
 
+    return TRUE;
 }
 
 static
@@ -1536,8 +1676,9 @@ cmsBool SetData(cmsContext ContextID, cmsIT8* it8, int nSet, int nField, const c
 {
     TABLE* t = GetTable(ContextID, it8);
 
-    if (!t->Data)
-        AllocateDataSet(ContextID, it8);
+    if (!t->Data) {
+        if (!AllocateDataSet(ContextID, it8)) return FALSE;
+    }
 
     if (!t->Data) return FALSE;
 
@@ -1674,11 +1815,11 @@ void WriteHeader(cmsContext ContextID, cmsIT8* it8, SAVESTREAM* fp)
                     break;
 
             case WRITE_HEXADECIMAL:
-                    Writef(ContextID, fp, "\t0x%X", atoi(p ->Value));
+                    Writef(ContextID, fp, "\t0x%X", satoi(p ->Value));
                     break;
 
             case WRITE_BINARY:
-                    Writef(ContextID, fp, "\t0x%B", atoi(p ->Value));
+                    Writef(ContextID, fp, "\t0b%s", satob(p ->Value));
                     break;
 
             case WRITE_PAIR:
@@ -1707,7 +1848,7 @@ void WriteDataFormat(cmsContext ContextID, SAVESTREAM* fp, cmsIT8* it8)
 
        WriteStr(ContextID, fp, "BEGIN_DATA_FORMAT\n");
        WriteStr(ContextID, fp, " ");
-       nSamples = atoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_FIELDS"));
+       nSamples = satoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_FIELDS"));
 
        for (i = 0; i < nSamples; i++) {
 
@@ -1730,7 +1871,7 @@ void WriteData(cmsContext ContextID, SAVESTREAM* fp, cmsIT8* it8)
 
        WriteStr (ContextID, fp, "BEGIN_DATA\n");
 
-       t->nPatches = atoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_SETS"));
+       t->nPatches = satoi(cmsIT8GetProperty(ContextID, it8, "NUMBER_OF_SETS"));
 
        for (i = 0; i < t-> nPatches; i++) {
 
@@ -1848,7 +1989,7 @@ cmsBool DataFormatSection(cmsContext ContextID, cmsIT8* it8)
                 return SynError(ContextID, it8, "Sample type expected");
             }
 
-            if (!SetDataFormat(ContextID, it8, iField, it8->id)) return FALSE;
+            if (!SetDataFormat(ContextID, it8, iField, StringPtr(it8->id))) return FALSE;
             iField++;
 
             InSymbol(ContextID, it8);
@@ -1881,8 +2022,9 @@ cmsBool DataSection (cmsContext ContextID, cmsIT8* it8)
     InSymbol(ContextID, it8);   // Eats "BEGIN_DATA"
     CheckEOLN(ContextID, it8);
 
-    if (!t->Data)
-        AllocateDataSet(ContextID, it8);
+    if (!t->Data) {
+        if (!AllocateDataSet(ContextID, it8)) return FALSE;
+    }
 
     while (it8->sy != SEND_DATA && it8->sy != SEOF)
     {
@@ -1894,11 +2036,28 @@ cmsBool DataSection (cmsContext ContextID, cmsIT8* it8)
 
         if (it8->sy != SEND_DATA && it8->sy != SEOF) {
 
+            switch (it8->sy)
+            {
+
+            // To keep very long data
+            case SIDENT:
+                if (!SetData(ContextID, it8, iSet, iField, StringPtr(it8->id)))
+                    return FALSE;
+                break;
+
+            case SSTRING:
+                if (!SetData(ContextID, it8, iSet, iField, StringPtr(it8->str)))
+                    return FALSE;
+                break;
+
+            default:
+
             if (!GetVal(ContextID, it8, Buffer, 255, "Sample data expected"))
                 return FALSE;
 
             if (!SetData(ContextID, it8, iSet, iField, Buffer))
                 return FALSE;
+            }
 
             iField++;
 
@@ -1954,7 +2113,7 @@ cmsBool HeaderSection(cmsContext ContextID, cmsIT8* it8)
 
 
         case SIDENT:
-            strncpy(VarName, it8->id, MAXID - 1);
+            strncpy(VarName, StringPtr(it8->id), MAXID - 1);
             VarName[MAXID - 1] = 0;
 
             if (!IsAvailableOnList(ContextID, it8->ValidKeywords, VarName, NULL, &Key)) {
@@ -2032,20 +2191,20 @@ cmsBool HeaderSection(cmsContext ContextID, cmsIT8* it8)
 
 
 static
-void ReadType(cmsContext ContextID, cmsIT8* it8, char* SheetTypePtr)
+void ReadType(cmsIT8* it8, char* SheetTypePtr)
 {
     cmsInt32Number cnt = 0;
 
     // First line is a very special case.
 
     while (isseparator(it8->ch))
-            NextCh(ContextID, it8);
+            NextCh(it8);
 
     while (it8->ch != '\r' && it8 ->ch != '\n' && it8->ch != '\t' && it8 -> ch != 0) {
 
         if (cnt++ < MAXSTR)
             *SheetTypePtr++= (char) it8 ->ch;
-        NextCh(ContextID, it8);
+        NextCh(it8);
     }
 
     *SheetTypePtr = 0;
@@ -2058,7 +2217,7 @@ cmsBool ParseIT8(cmsContext ContextID, cmsIT8* it8, cmsBool nosheet)
     char* SheetTypePtr = it8 ->Tab[0].SheetType;
 
     if (nosheet == 0) {
-        ReadType(ContextID, it8, SheetTypePtr);
+        ReadType(it8, SheetTypePtr);
     }
 
     InSymbol(ContextID, it8);
@@ -2093,12 +2252,12 @@ cmsBool ParseIT8(cmsContext ContextID, cmsIT8* it8, cmsBool nosheet)
                                     // May be a type sheet or may be a prop value statement. We cannot use insymbol in
                                     // this special case...
                                      while (isseparator(it8->ch))
-                                         NextCh(ContextID, it8);
+                                         NextCh(it8);
 
                                      // If a newline is found, then this is a type string
                                     if (it8 ->ch == '\n' || it8->ch == '\r') {
 
-                                         cmsIT8SetSheetType(ContextID, it8, it8 ->id);
+                                         cmsIT8SetSheetType(ContextID, it8, StringPtr(it8 ->id));
                                          InSymbol(ContextID, it8);
                                     }
                                     else
@@ -2110,7 +2269,7 @@ cmsBool ParseIT8(cmsContext ContextID, cmsIT8* it8, cmsBool nosheet)
                                 else
                                     // Validate quoted strings
                                     if (it8 ->sy == SSTRING) {
-                                        cmsIT8SetSheetType(ContextID, it8, it8 ->str);
+                                        cmsIT8SetSheetType(ContextID, it8, StringPtr(it8 ->str));
                                         InSymbol(ContextID, it8);
                                     }
                            }
@@ -2164,71 +2323,53 @@ void CookPointers(cmsContext ContextID, cmsIT8* it8)
         if (cmsstrcasecmp(Fld, "SAMPLE_ID") == 0) {
 
             t -> SampleID = idField;
-
-            for (i=0; i < t -> nPatches; i++) {
-
-                char *Data = GetData(ContextID, it8, i, idField);
-                if (Data) {
-                    char Buffer[256];
-
-                    strncpy(Buffer, Data, 255);
-                    Buffer[255] = 0;
-
-                    if (strlen(Buffer) <= strlen(Data))
-                        strcpy(Data, Buffer);
-                    else
-                        SetData(ContextID, it8, i, idField, Buffer);
-
-                }
-            }
-
         }
 
         // "LABEL" is an extension. It keeps references to forward tables
 
-        if ((cmsstrcasecmp(Fld, "LABEL") == 0) || Fld[0] == '$' ) {
+        if ((cmsstrcasecmp(Fld, "LABEL") == 0) || Fld[0] == '$') {
 
-                    // Search for table references...
-                    for (i=0; i < t -> nPatches; i++) {
+            // Search for table references...
+            for (i = 0; i < t->nPatches; i++) {
 
-                            char *Label = GetData(ContextID, it8, i, idField);
+                char *Label = GetData(ContextID, it8, i, idField);
 
-                            if (Label) {
+                if (Label) {
 
-                                cmsUInt32Number k;
+                    cmsUInt32Number k;
 
-                                // This is the label, search for a table containing
-                                // this property
+                    // This is the label, search for a table containing
+                    // this property
 
-                                for (k=0; k < it8 ->TablesCount; k++) {
+                    for (k = 0; k < it8->TablesCount; k++) {
 
-                                    TABLE* Table = it8 ->Tab + k;
-                                    KEYVALUE* p;
+                        TABLE* Table = it8->Tab + k;
+                        KEYVALUE* p;
 
-                                    if (IsAvailableOnList(ContextID, Table->HeaderList, Label, NULL, &p)) {
+                        if (IsAvailableOnList(ContextID, Table->HeaderList, Label, NULL, &p)) {
 
-                                        // Available, keep type and table
-                                        char Buffer[256];
+                            // Available, keep type and table
+                            char Buffer[256];
 
-                                        char *Type  = p ->Value;
-                                        int  nTable = (int) k;
+                            char* Type = p->Value;
+                            int  nTable = (int)k;
 
-                                        snprintf(Buffer, 255, "%s %d %s", Label, nTable, Type );
-
-                                        SetData(ContextID, it8, i, idField, Buffer);
-                                    }
-                                }
-
-
-                            }
+                            snprintf(Buffer, 255, "%s %d %s", Label, nTable, Type);
 
+                            SetData(ContextID, it8, i, idField, Buffer);
+                        }
                     }
 
 
+                }
+
+            }
+
+
         }
 
-    }
-    }
+	}
+	}
 
     it8 ->nTable = nOldTable;
 }
@@ -2237,7 +2378,7 @@ void CookPointers(cmsContext ContextID, cmsIT8* it8)
 // that should be something like some printable characters plus a \n
 // returns 0 if this is not like a CGATS, or an integer otherwise. This integer is the number of words in first line?
 static
-int IsMyBlock(cmsContext ContextID, const cmsUInt8Number* Buffer, cmsUInt32Number n)
+int IsMyBlock(const cmsUInt8Number* Buffer, cmsUInt32Number n)
 {
     int words = 1, space = 0, quot = 0;
     cmsUInt32Number i;
@@ -2276,7 +2417,7 @@ int IsMyBlock(cmsContext ContextID, const cmsUInt8Number* Buffer, cmsUInt32Numbe
 
 
 static
-cmsBool IsMyFile(cmsContext ContextID, const char* FileName)
+cmsBool IsMyFile(const char* FileName)
 {
    FILE *fp;
    cmsUInt32Number Size;
@@ -2295,7 +2436,7 @@ cmsBool IsMyFile(cmsContext ContextID, const char* FileName)
 
    Ptr[Size] = '\0';
 
-   return IsMyBlock(ContextID, Ptr, Size);
+   return IsMyBlock(Ptr, Size);
 }
 
 // ---------------------------------------------------------- Exported routines
@@ -2310,7 +2451,7 @@ cmsHANDLE  CMSEXPORT cmsIT8LoadFromMem(cmsContext ContextID, const void *Ptr, cm
     _cmsAssert(Ptr != NULL);
     _cmsAssert(len != 0);
 
-    type = IsMyBlock(ContextID, (const cmsUInt8Number*)Ptr, len);
+    type = IsMyBlock((const cmsUInt8Number*)Ptr, len);
     if (type == 0) return NULL;
 
     hIT8 = cmsIT8Alloc(ContextID);
@@ -2318,6 +2459,11 @@ cmsHANDLE  CMSEXPORT cmsIT8LoadFromMem(cmsContext ContextID, const void *Ptr, cm
 
     it8 = (cmsIT8*) hIT8;
     it8 ->MemoryBlock = (char*) _cmsMalloc(ContextID, len + 1);
+    if (it8->MemoryBlock == NULL)
+    {
+        cmsIT8Free(ContextID, hIT8);
+        return FALSE;
+    }
 
     strncpy(it8 ->MemoryBlock, (const char*) Ptr, len);
     it8 ->MemoryBlock[len] = 0;
@@ -2352,7 +2498,7 @@ cmsHANDLE  CMSEXPORT cmsIT8LoadFromFile(cmsContext ContextID, const char* cFileN
 
      _cmsAssert(cFileName != NULL);
 
-     type = IsMyFile(ContextID, cFileName);
+     type = IsMyFile(cFileName);
      if (type == 0) return NULL;
 
      hIT8 = cmsIT8Alloc(ContextID);
@@ -2426,14 +2572,17 @@ cmsUInt32Number CMSEXPORT cmsIT8EnumProperties(cmsContext ContextID, cmsHANDLE h
 
 
     Props = (char **) AllocChunk(ContextID, it8, sizeof(char *) * n);
+    if (Props != NULL) {
+
+	// Pass#2 - Fill pointers
+	n = 0;
+	for (p = t->HeaderList; p != NULL; p = p->Next) {
+		Props[n++] = p->Keyword;
+	}
 
-    // Pass#2 - Fill pointers
-    n = 0;
-    for (p = t -> HeaderList;  p != NULL; p = p->Next) {
-        Props[n++] = p -> Keyword;
     }
-
     *PropertyNames = Props;
+
     return n;
 }
 
@@ -2465,12 +2614,14 @@ cmsUInt32Number CMSEXPORT cmsIT8EnumPropertyMulti(cmsContext ContextID, cmsHANDL
 
 
     Props = (const char **) AllocChunk(ContextID, it8, sizeof(char *) * n);
+    if (Props != NULL) {
 
-    // Pass#2 - Fill pointers
-    n = 0;
-    for (tmp = p;  tmp != NULL; tmp = tmp->NextSubkey) {
-        if(tmp->Subkey != NULL)
-            Props[n++] = p ->Subkey;
+        // Pass#2 - Fill pointers
+        n = 0;
+        for (tmp = p; tmp != NULL; tmp = tmp->NextSubkey) {
+            if (tmp->Subkey != NULL)
+                Props[n++] = p->Subkey;
+        }
     }
 
     *SubpropertyNames = Props;
@@ -2569,7 +2720,7 @@ cmsFloat64Number CMSEXPORT cmsIT8GetDataRowColDbl(cmsContext ContextID, cmsHANDL
 
     if (Buffer == NULL) return 0.0;
 
-    return ParseFloatNumber(ContextID, Buffer);
+    return ParseFloatNumber(Buffer);
 }
 
 
@@ -2624,7 +2775,7 @@ cmsFloat64Number CMSEXPORT cmsIT8GetDataDbl(cmsContext ContextID, cmsHANDLE  it8
 
     Buffer = cmsIT8GetData(ContextID, it8, cPatch, cSample);
 
-    return ParseFloatNumber(ContextID, Buffer);
+    return ParseFloatNumber(Buffer);
 }
 
 
@@ -2646,8 +2797,12 @@ cmsBool CMSEXPORT cmsIT8SetData(cmsContext ContextID, cmsHANDLE hIT8, const char
 
     if (t-> nPatches == 0) {
 
-        AllocateDataFormat(ContextID, it8);
-        AllocateDataSet(ContextID, it8);
+        if (!AllocateDataFormat(ContextID, it8))
+            return FALSE;
+
+        if (!AllocateDataSet(ContextID, it8))
+            return FALSE;
+
         CookPointers(ContextID, it8);
     }
 
